import math
from coramin.relaxations.relaxations_base import (
    BaseRelaxationData,
    BasePWRelaxationData,
)
import pyomo.environ as pe
from pyomo.common.timing import HierarchicalTimer
from pyomo.core.base.block import _BlockData
from pyomo.contrib.appsi.base import (
    Results,
    PersistentSolver,
    Solver,
    MIPSolverConfig,
    TerminationCondition,
    SolutionLoaderBase,
    UpdateConfig,
)
from pyomo.contrib import appsi
from typing import Tuple, Optional, MutableMapping, Sequence
from pyomo.common.config import (
    ConfigValue, NonNegativeInt, PositiveFloat, PositiveInt, NonNegativeFloat, InEnum
)
import logging
from coramin.relaxations.auto_relax import relax
from coramin.relaxations.iterators import relaxation_data_objects
from coramin.utils.coramin_enums import RelaxationSide, Effort, EigenValueBounder
from coramin.domain_reduction import push_integers, pop_integers, collect_vars_to_tighten, perform_obbt
import time
from pyomo.core.base.var import _GeneralVarData
from pyomo.core.base.objective import _GeneralObjectiveData
from coramin.utils.pyomo_utils import get_objective
from pyomo.common.collections.component_set import ComponentSet
from pyomo.common.modeling import unique_component_name
from pyomo.common.errors import InfeasibleConstraintException
from pyomo.contrib.fbbt.fbbt import BoundsManager
import numpy as np


logger = logging.getLogger(__name__)


class MultiTreeConfig(MIPSolverConfig):
    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super(MultiTreeConfig, self).__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        self.declare("solver_output_logger", ConfigValue())
        self.declare("log_level", ConfigValue(domain=NonNegativeInt))
        self.declare("feasibility_tolerance", ConfigValue(domain=PositiveFloat))
        self.declare("abs_gap", ConfigValue(domain=PositiveFloat))
        self.declare("max_partitions_per_iter", ConfigValue(domain=PositiveInt))
        self.declare("max_iter", ConfigValue(domain=NonNegativeInt))
        self.declare("root_obbt_max_iter", ConfigValue(domain=NonNegativeInt))
        self.declare("show_obbt_progress_bar", ConfigValue(domain=bool))
        self.declare("integer_tolerance", ConfigValue(domain=PositiveFloat))
        self.declare("small_coef", ConfigValue(domain=NonNegativeFloat))
        self.declare("large_coef", ConfigValue(domain=NonNegativeFloat))
        self.declare("safety_tol", ConfigValue(domain=NonNegativeFloat))
        self.declare("convexity_effort", ConfigValue(domain=InEnum(Effort)))

        self.solver_output_logger = logger
        self.log_level = logging.INFO
        self.feasibility_tolerance = 1e-6
        self.integer_tolerance = 1e-4
        self.time_limit = 600
        self.abs_gap = 1e-4
        self.mip_gap = 0.001
        self.max_partitions_per_iter = 5
        self.max_iter = 1000
        self.root_obbt_max_iter = 1000
        self.show_obbt_progress_bar = False
        self.small_coef = 1e-10
        self.large_coef = 1e5
        self.safety_tol = 1e-10
        self.convexity_effort = Effort.high


def _is_problem_definitely_convex(m: _BlockData) -> bool:
    res = True
    for r in relaxation_data_objects(m, descend_into=True, active=True):
        if r.relaxation_side == RelaxationSide.BOTH:
            res = False
            break
        elif r.relaxation_side == RelaxationSide.UNDER and not r.is_rhs_convex():
            res = False
            break
        elif r.relaxation_side == RelaxationSide.OVER and not r.is_rhs_concave():
            res = False
            break
    return res


class MultiTreeResults(Results):
    def __init__(self):
        super().__init__()
        self.wallclock_time = None


class MultiTreeSolutionLoader(SolutionLoaderBase):
    def __init__(self, primals: MutableMapping):
        self._primals = primals

    def get_primals(
        self, vars_to_load: Optional[Sequence[_GeneralVarData]] = None
    ) -> MutableMapping[_GeneralVarData, float]:
        if vars_to_load is None:
            return pe.ComponentMap(self._primals.items())
        else:
            primals = pe.ComponentMap()
            for v in vars_to_load:
                primals[v] = self._primals[v]
        return primals


class MultiTree(Solver):
    def __init__(self, mip_solver: PersistentSolver, nlp_solver: PersistentSolver):
        super(MultiTree, self).__init__()
        self._config = MultiTreeConfig()
        self.mip_solver: PersistentSolver = mip_solver
        self.nlp_solver: PersistentSolver = nlp_solver
        self._original_model: Optional[_BlockData] = None
        self._relaxation: Optional[_BlockData] = None
        self._nlp: Optional[_BlockData] = None
        self._start_time: Optional[float] = None
        self._incumbent: Optional[pe.ComponentMap] = None
        self._best_feasible_objective: Optional[float] = None
        self._best_objective_bound: Optional[float] = None
        self._objective: Optional[_GeneralObjectiveData] = None
        self._relaxation_objects: Optional[Sequence[BaseRelaxationData]] = None
        self._stop: Optional[TerminationCondition] = None
        self._discrete_vars: Optional[Sequence[_GeneralVarData]] = None
        self._rel_to_nlp_map: Optional[MutableMapping] = None
        self._nlp_to_orig_map: Optional[MutableMapping] = None
        self._nlp_tightener: Optional[appsi.fbbt.IntervalTightener] = None
        self._iter: int = 0

    def _re_init(self):
        self._original_model: Optional[_BlockData] = None
        self._relaxation: Optional[_BlockData] = None
        self._nlp: Optional[_BlockData] = None
        self._start_time: Optional[float] = None
        self._incumbent: Optional[pe.ComponentMap] = None
        self._best_feasible_objective: Optional[float] = None
        self._best_objective_bound: Optional[float] = None
        self._objective: Optional[_GeneralObjectiveData] = None
        self._relaxation_objects: Optional[Sequence[BaseRelaxationData]] = None
        self._stop: Optional[TerminationCondition] = None
        self._discrete_vars: Optional[Sequence[_GeneralVarData]] = None
        self._rel_to_nlp_map: Optional[MutableMapping] = None
        self._nlp_to_orig_map: Optional[MutableMapping] = None
        self._nlp_tightener: Optional[appsi.fbbt.IntervalTightener] = None
        self._iter: int = 0

    def available(self):
        if (
            self.mip_solver.available() == Solver.Availability.FullLicense
            and self.nlp_solver.available() == Solver.Availability.FullLicense
        ):
            return Solver.Availability.FullLicense
        elif self.mip_solver.available() == Solver.Availability.FullLicense:
            return self.nlp_solver.available()
        else:
            return self.mip_solver.available()

    def version(self) -> Tuple:
        return 0, 1, 0

    @property
    def config(self) -> MultiTreeConfig:
        return self._config

    @config.setter
    def config(self, val: MultiTreeConfig):
        self._config = val

    @property
    def symbol_map(self):
        raise NotImplementedError("This solver does not have a symbol map")

    def _should_terminate(self) -> Tuple[bool, Optional[TerminationCondition]]:
        if self._elapsed_time >= self.config.time_limit:
            return True, TerminationCondition.maxTimeLimit
        if self._iter >= self.config.max_iter:
            return True, TerminationCondition.maxIterations
        if self._stop is not None:
            return True, self._stop
        primal_bound = self._get_primal_bound()
        dual_bound = self._get_dual_bound()
        if self._objective.sense == pe.minimize:
            assert primal_bound >= dual_bound - 1e-6*max(abs(primal_bound), abs(dual_bound)) - 1e-6
        else:
            assert primal_bound <= dual_bound + 1e-6*max(abs(primal_bound), abs(dual_bound)) + 1e-6
        abs_gap, rel_gap = self._get_abs_and_rel_gap()
        if abs_gap <= self.config.abs_gap:
            return True, TerminationCondition.optimal
        if rel_gap <= self.config.mip_gap:
            return True, TerminationCondition.optimal
        return False, TerminationCondition.unknown

    def _get_results(self, termination_condition: TerminationCondition) -> MultiTreeResults:
        res = MultiTreeResults()
        res.termination_condition = termination_condition
        res.best_feasible_objective = self._best_feasible_objective
        res.best_objective_bound = self._best_objective_bound
        if self._best_feasible_objective is not None:
            res.solution_loader = MultiTreeSolutionLoader(self._incumbent)
        res.wallclock_time = self._elapsed_time

        if self.config.load_solution:
            if res.best_feasible_objective is not None:
                if res.termination_condition != TerminationCondition.optimal:
                    logger.warning('Loading a feasible but potentially sub-optimal '
                                   'solution. Please check the termination condition.')
                res.solution_loader.load_vars()
            else:
                raise RuntimeError('No feasible solution was found. Please '
                                   'set opt.config.load_solution=False and check the '
                                   'termination condition before loading a solution.')

        return res

    def _get_primal_bound(self) -> float:
        if self._best_feasible_objective is None:
            if self._objective.sense == pe.minimize:
                primal_bound = math.inf
            else:
                primal_bound = -math.inf
        else:
            primal_bound = self._best_feasible_objective
        return primal_bound

    def _get_dual_bound(self) -> float:
        if self._best_objective_bound is None:
            if self._objective.sense == pe.minimize:
                dual_bound = -math.inf
            else:
                dual_bound = math.inf
        else:
            dual_bound = self._best_objective_bound
        return dual_bound

    def _get_abs_and_rel_gap(self):
        primal_bound = self._get_primal_bound()
        dual_bound = self._get_dual_bound()
        abs_gap = abs(primal_bound - dual_bound)
        if abs_gap == 0:
            rel_gap = 0
        elif primal_bound == 0:
            rel_gap = math.inf
        elif math.isinf(abs_gap):
            rel_gap = math.inf
        else:
            rel_gap = abs_gap / abs(primal_bound)
        return abs_gap, rel_gap

    def _get_constr_violation(self):
        viol_list = list()
        if len(self._relaxation_objects) == 0:
            return 0
        for b in self._relaxation_objects:
            any_none = False
            for v in b.get_rhs_vars():
                if v.value is None:
                    any_none = True
                    break
            if any_none:
                viol_list.append(math.inf)
                break
            else:
                viol_list.append(b.get_deviation())
        return max(viol_list)

    def _log(self, header=False, num_lb_improved=0, num_ub_improved=0,
             avg_lb_improvement=0, avg_ub_improvement=0, rel_termination=None,
             nlp_termination=None, constr_viol=None):
        logger = self.config.solver_output_logger
        log_level = self.config.log_level
        if header:
            msg = (
                f"{'Iter':<6}{'Primal Bd':<12}{'Dual Bd':<12}{'Abs Gap':<9}"
                f"{'% Gap':<7}{'CnstrVio':<10}{'Time':<6}{'NLP Term':<10}"
                f"{'Rel Term':<10}{'#LBs':<6}{'#UBs':<6}{'Avg LB':<9}"
                f"{'Avg UB':<9}"
            )
            logger.log(log_level, msg)
            if self.config.stream_solver:
                print(msg)
        else:
            if rel_termination is None:
                rel_termination = '-'
            else:
                rel_termination = str(rel_termination.name)
            if nlp_termination is None:
                nlp_termination = '-'
            else:
                nlp_termination = str(nlp_termination.name)
            rel_termination = rel_termination[:9]
            nlp_termination = nlp_termination[:9]
            primal_bound = self._get_primal_bound()
            dual_bound = self._get_dual_bound()
            abs_gap, rel_gap = self._get_abs_and_rel_gap()
            if constr_viol is None:
                constr_viol = '-'
            else:
                constr_viol = f'{constr_viol:<10.1e}'
            elapsed_time = self._elapsed_time
            if elapsed_time < 100:
                elapsed_time_str = f'{elapsed_time:<6.2f}'
            else:
                elapsed_time_str = f'{round(elapsed_time):<6d}'
            percent_gap = rel_gap*100
            if math.isinf(percent_gap):
                percent_gap_str = f'{percent_gap:<7.2f}'
            elif percent_gap >= 100:
                percent_gap_str = f'{round(percent_gap):<7d}'
            else:
                percent_gap_str = f'{percent_gap:<7.2f}'
            msg = (
                f"{self._iter:<6}{primal_bound:<12.3e}{dual_bound:<12.3e}"
                f"{abs_gap:<9.1e}{percent_gap_str:<7}{constr_viol:<10}"
                f"{elapsed_time_str:<6}{nlp_termination:<10}"
                f"{rel_termination:<10}{num_lb_improved:<6}"
                f"{num_ub_improved:<6}{avg_lb_improvement:<9.1e}"
                f"{avg_ub_improvement:<9.1e}"
            )
            logger.log(log_level, msg)
            if self.config.stream_solver:
                print(msg)

    def _update_dual_bound(self, res: Results):
        if res.best_objective_bound is not None:
            if self._objective.sense == pe.minimize:
                if (
                    self._best_objective_bound is None
                    or res.best_objective_bound > self._best_objective_bound
                ):
                    self._best_objective_bound = res.best_objective_bound
            else:
                if (
                    self._best_objective_bound is None
                    or res.best_objective_bound < self._best_objective_bound
                ):
                    self._best_objective_bound = res.best_objective_bound

        if res.best_feasible_objective is not None:
            max_viol = self._get_constr_violation()
            if max_viol > self.config.feasibility_tolerance:
                all_cons_satisfied = False
            else:
                all_cons_satisfied = True
            if all_cons_satisfied:
                for v in self._discrete_vars:
                    if v.value is None:
                        assert v.stale
                        continue
                    if not math.isclose(v.value, round(v.value), rel_tol=self.config.integer_tolerance, abs_tol=self.config.integer_tolerance):
                        all_cons_satisfied = False
                        break
            if all_cons_satisfied:
                for rel_v, nlp_v in self._rel_to_nlp_map.items():
                    if rel_v.value is None:
                        assert rel_v.stale
                        if rel_v.has_lb() and rel_v.has_ub() and math.isclose(rel_v.lb, rel_v.ub, rel_tol=self.config.feasibility_tolerance, abs_tol=self.config.feasibility_tolerance):
                            nlp_v.value = 0.5*(rel_v.lb + rel_v.ub)
                        else:
                            nlp_v.value = None
                    else:
                        nlp_v.set_value(rel_v.value, skip_validation=True)
                self._update_primal_bound(res)

    def _update_primal_bound(self, res: Results):
        should_update = False
        if res.best_feasible_objective is not None:
            if self._objective.sense == pe.minimize:
                if (
                    self._best_feasible_objective is None
                    or res.best_feasible_objective < self._best_feasible_objective
                ):
                    should_update = True
            else:
                if (
                    self._best_feasible_objective is None
                    or res.best_feasible_objective > self._best_feasible_objective
                ):
                    should_update = True

        if should_update:
            self._best_feasible_objective = res.best_feasible_objective
            self._incumbent = pe.ComponentMap()
            for nlp_v, orig_v in self._nlp_to_orig_map.items():
                self._incumbent[orig_v] = nlp_v.value

    def _solve_nlp_with_fixed_vars(
        self,
        integer_var_values: MutableMapping[_GeneralVarData, float],
        rhs_var_bounds: MutableMapping[_GeneralVarData, Tuple[float, float]],
    ) -> Results:
        self._iter += 1

        bm = BoundsManager(self._nlp)
        bm.save_bounds()

        fixed_vars = list()
        for v in self._discrete_vars:
            if v.fixed:
                continue
            val = integer_var_values[v]
            assert math.isclose(val, round(val), rel_tol=self.config.integer_tolerance, abs_tol=self.config.integer_tolerance)
            val = round(val)
            nlp_v = self._rel_to_nlp_map[v]
            orig_v = self._nlp_to_orig_map[nlp_v]
            nlp_v.fix(val)
            orig_v.fix(val)
            fixed_vars.append(nlp_v)
            fixed_vars.append(orig_v)

        for v, (v_lb, v_ub) in rhs_var_bounds.items():
            if v.fixed:
                continue
            nlp_v = self._rel_to_nlp_map[v]
            nlp_v.setlb(v_lb)
            nlp_v.setub(v_ub)

        nlp_res = Results()

        active_constraints = list()
        for c in ComponentSet(
            self._nlp.component_data_objects(
                pe.Constraint, active=True, descend_into=True
            )
        ):
            active_constraints.append(c)

        try:
            self._nlp_tightener.perform_fbbt(self._nlp)
            proven_infeasible = False
        except InfeasibleConstraintException:
            # the original NLP may still be feasible
            proven_infeasible = True

        if proven_infeasible:
            any_unfixed_vars = False
            for v in self._original_model.component_data_objects(
                pe.Var, descend_into=True
            ):
                if not v.fixed:
                    any_unfixed_vars = True
                    break
            if any_unfixed_vars:
                self.nlp_solver.config.time_limit = self._remaining_time
                self._original_model.pprint()
                nlp_res = self.nlp_solver.solve(self._original_model)
                if nlp_res.best_feasible_objective is not None:
                    nlp_res.solution_loader.load_vars()
                    for nlp_v, orig_v in self._nlp_to_orig_map.items():
                        nlp_v.set_value(orig_v.value, skip_validation=True)
                else:
                    nlp_res = Results()
                    nlp_res.termination_condition = TerminationCondition.infeasible
        else:
            for v in ComponentSet(
                self._nlp.component_data_objects(pe.Var, descend_into=True)
            ):
                if v.fixed:
                    continue
                if v.has_lb() and v.has_ub():
                    if math.isclose(v.lb, v.ub, rel_tol=self.config.feasibility_tolerance, abs_tol=self.config.feasibility_tolerance):
                        v.fix(0.5 * (v.lb + v.ub))
                        fixed_vars.append(v)
                    else:
                        v.value = 0.5 * (v.lb + v.ub)

            any_unfixed_vars = False
            for v in self._nlp.component_data_objects(
                pe.Var, descend_into=True
            ):
                if not v.fixed:
                    any_unfixed_vars = True
                    break

            if any_unfixed_vars:
                self.nlp_solver.config.time_limit = self._remaining_time
                self.nlp_solver.config.load_solution = False
                nlp_res = self.nlp_solver.solve(self._nlp)
                if nlp_res.best_feasible_objective is not None:
                    nlp_res.solution_loader.load_vars()
                else:
                    self.nlp_solver.config.time_limit = self._remaining_time
                    nlp_res = self.nlp_solver.solve(self._original_model)
                    if nlp_res.best_feasible_objective is not None:
                        nlp_res.solution_loader.load_vars()
                        for nlp_v, orig_v in self._nlp_to_orig_map.items():
                            nlp_v.value = orig_v.value
            else:
                nlp_obj = get_objective(self._nlp)
                # there should not be any active constraints
                # they should all have been deactivated by FBBT
                for c in active_constraints:
                    assert not c.active
                nlp_res.termination_condition = TerminationCondition.optimal
                nlp_res.best_feasible_objective = pe.value(nlp_obj)
                nlp_res.best_objective_bound = nlp_res.best_feasible_objective
                nlp_res.solution_loader = MultiTreeSolutionLoader(pe.ComponentMap((v, v.value) for v in self._nlp.component_data_objects(pe.Var, descend_into=True)))

        self._update_primal_bound(nlp_res)
        self._log(header=False, nlp_termination=nlp_res.termination_condition)

        for v in fixed_vars:
            v.unfix()

        bm.pop_bounds()

        for c in active_constraints:
            c.activate()

        return nlp_res

    def _solve_relaxation(self) -> Results:
        self._iter += 1
        self.mip_solver.config.time_limit = self._remaining_time
        self.mip_solver.config.load_solution = False
        rel_res = self.mip_solver.solve(self._relaxation)

        if rel_res.best_feasible_objective is not None:
            rel_res.solution_loader.load_vars()

        self._update_dual_bound(rel_res)
        self._log(header=False, rel_termination=rel_res.termination_condition, constr_viol=self._get_constr_violation())
        if rel_res.termination_condition not in {
            TerminationCondition.optimal,
            TerminationCondition.maxTimeLimit,
            TerminationCondition.maxIterations,
            TerminationCondition.objectiveLimit,
            TerminationCondition.interrupted,
        }:
            self._stop = rel_res.termination_condition
        return rel_res

    def _partition_helper(self):
        dev_list = list()

        err = False

        for b in self._relaxation_objects:
            for v in b.get_rhs_vars():
                if not v.has_lb() or not v.has_ub():
                    logger.error(
                        'The multitree algorithm is not guaranteed to converge '
                        'for problems with unbounded variables. Please bound all '
                        'variables.')
                    self._stop = TerminationCondition.error
                    err = True
                    break
            if err:
                break

            aux_val = b.get_aux_var().value
            rhs_val = pe.value(b.get_rhs_expr())
            if (
                aux_val > rhs_val + self.config.feasibility_tolerance
                and b.relaxation_side in {RelaxationSide.BOTH, RelaxationSide.OVER}
                and not b.is_rhs_concave()
            ):
                dev_list.append((b, aux_val - rhs_val))
            elif (
                aux_val < rhs_val - self.config.feasibility_tolerance
                and b.relaxation_side in {RelaxationSide.BOTH, RelaxationSide.UNDER}
                and not b.is_rhs_convex()
            ):
                dev_list.append((b, rhs_val - aux_val))

        if not err:
            dev_list.sort(key=lambda x: x[1], reverse=True)

            for b, dev in dev_list[: self.config.max_partitions_per_iter]:
                b.add_partition_point()
                b.rebuild()

    def _oa_cut_helper(self, tol):
        new_con_list = list()
        for b in self._relaxation_objects:
            new_con = b.add_cut(
                keep_cut=True, check_violation=True, feasibility_tol=tol
            )
            if new_con is not None:
                new_con_list.append(new_con)
        self.mip_solver.add_constraints(new_con_list)
        return new_con_list

    def _add_oa_cuts(self, tol, max_iter) -> Results:
        original_update_config: UpdateConfig = self.mip_solver.update_config()

        self.mip_solver.update()

        self.mip_solver.update_config.update_params = False
        self.mip_solver.update_config.update_vars = False
        self.mip_solver.update_config.update_objective = False
        self.mip_solver.update_config.update_constraints = False
        self.mip_solver.update_config.check_for_new_objective = False
        self.mip_solver.update_config.check_for_new_or_removed_constraints = False
        self.mip_solver.update_config.check_for_new_or_removed_vars = False
        self.mip_solver.update_config.check_for_new_or_removed_params = True
        self.mip_solver.update_config.treat_fixed_vars_as_params = True
        self.mip_solver.update_config.update_named_expressions = False

        last_res = None

        for _iter in range(max_iter):
            if self._should_terminate()[0]:
                break

            rel_res = self._solve_relaxation()
            if rel_res.best_feasible_objective is not None:
                last_res = Results()
                last_res.best_feasible_objective = rel_res.best_feasible_objective
                last_res.best_objective_bound = rel_res.best_objective_bound
                last_res.termination_condition = rel_res.termination_condition
                last_res.solution_loader = MultiTreeSolutionLoader(
                    rel_res.solution_loader.get_primals(
                        vars_to_load=self._discrete_vars
                    )
                )

            if self._should_terminate()[0]:
                break

            new_con_list = self._oa_cut_helper(tol=tol)
            if len(new_con_list) == 0:
                break

        self.mip_solver.update_config.update_params = (
            original_update_config.update_params
        )
        self.mip_solver.update_config.update_vars = original_update_config.update_vars
        self.mip_solver.update_config.update_objective = (
            original_update_config.update_objective
        )
        self.mip_solver.update_config.update_constraints = (
            original_update_config.update_constraints
        )
        self.mip_solver.update_config.check_for_new_objective = (
            original_update_config.check_for_new_objective
        )
        self.mip_solver.update_config.check_for_new_or_removed_constraints = (
            original_update_config.check_for_new_or_removed_constraints
        )
        self.mip_solver.update_config.check_for_new_or_removed_vars = (
            original_update_config.check_for_new_or_removed_vars
        )
        self.mip_solver.update_config.check_for_new_or_removed_params = (
            original_update_config.check_for_new_or_removed_params
        )
        self.mip_solver.update_config.treat_fixed_vars_as_params = (
            original_update_config.treat_fixed_vars_as_params
        )
        self.mip_solver.update_config.update_named_expressions = (
            original_update_config.update_named_expressions
        )

        if last_res is None:
            last_res = Results()

        return last_res

    def _construct_nlp(self):
        all_vars = list(
            ComponentSet(
                self._original_model.component_data_objects(pe.Var, descend_into=True)
            )
        )
        tmp_name = unique_component_name(self._original_model, "all_vars")
        setattr(self._original_model, tmp_name, all_vars)

        # this has to be 0 because the Multitree solver cannot use alpha-bb relaxations
        max_vars_per_alpha_bb = 0
        max_eigenvalue_for_alpha_bb = 0

        if self.config.convexity_effort == Effort.none:
            perform_expression_simplification = False
            use_alpha_bb = False
            eigenvalue_bounder = EigenValueBounder.Gershgorin
            eigenvalue_opt = None
        elif self.config.convexity_effort <= Effort.low:
            perform_expression_simplification = False
            use_alpha_bb = True
            eigenvalue_bounder = EigenValueBounder.Gershgorin
            eigenvalue_opt = None
        elif self.config.convexity_effort <= Effort.medium:
            perform_expression_simplification = True
            use_alpha_bb = True
            eigenvalue_bounder = EigenValueBounder.GershgorinWithSimplification
            eigenvalue_opt = None
        elif self.config.convexity_effort <= Effort.high:
            perform_expression_simplification = True
            use_alpha_bb = True
            eigenvalue_bounder = EigenValueBounder.LinearProgram
            eigenvalue_opt = self.mip_solver.__class__()
            eigenvalue_opt.config = self.mip_solver.config()
            # TODO: need to update the solver options
        else:
            perform_expression_simplification = True
            use_alpha_bb = True
            eigenvalue_bounder = EigenValueBounder.Global
            mip_solver = self.mip_solver.__class__()
            mip_solver.config = self.mip_solver.config()
            nlp_solver = self.nlp_solver.__class__()
            nlp_solver.config = self.nlp_solver.config()
            eigenvalue_opt = MultiTree(mip_solver=mip_solver, nlp_solver=nlp_solver)
            eigenvalue_opt.config = self.config()
            eigenvalue_opt.config.convexity_effort = min(self.config.convexity_effort, Effort.medium)

        self._nlp = relax(
            model=self._original_model,
            in_place=False,
            use_fbbt=True,
            fbbt_options={"deactivate_satisfied_constraints": True, "max_iter": 5},
            perform_expression_simplification=perform_expression_simplification,
            use_alpha_bb=use_alpha_bb,
            eigenvalue_bounder=eigenvalue_bounder,
            eigenvalue_opt=eigenvalue_opt,
            max_vars_per_alpha_bb=max_vars_per_alpha_bb,
            max_eigenvalue_for_alpha_bb=max_eigenvalue_for_alpha_bb,
        )
        new_vars = getattr(self._nlp, tmp_name)
        self._nlp_to_orig_map = pe.ComponentMap(zip(new_vars, all_vars))
        delattr(self._original_model, tmp_name)
        delattr(self._nlp, tmp_name)

        for b in relaxation_data_objects(self._nlp, descend_into=True, active=True):
            b.rebuild(build_nonlinear_constraint=True)

    def _construct_relaxation(self):
        all_vars = list(
            ComponentSet(
                self._nlp.component_data_objects(pe.Var, descend_into=True)
            )
        )
        tmp_name = unique_component_name(self._nlp, "all_vars")
        setattr(self._nlp, tmp_name, all_vars)
        self._relaxation = self._nlp.clone()
        new_vars = getattr(self._relaxation, tmp_name)
        self._rel_to_nlp_map = pe.ComponentMap(zip(new_vars, all_vars))
        delattr(self._nlp, tmp_name)
        delattr(self._relaxation, tmp_name)

        for b in relaxation_data_objects(self._relaxation, descend_into=True, active=True):
            b.small_coef = self.config.small_coef
            b.large_coef = self.config.large_coef
            b.safety_tol = self.config.safety_tol
            b.rebuild()

    def _get_nlp_specs_from_rel(self):
        integer_var_values = pe.ComponentMap()
        for v in self._discrete_vars:
            integer_var_values[v] = v.value
        rhs_var_bounds = pe.ComponentMap()
        for r in self._relaxation_objects:
            if not isinstance(r, BasePWRelaxationData):
                continue
            any_unbounded_vars = False
            for v in r.get_rhs_vars():
                if not v.has_lb() or not v.has_ub():
                    any_unbounded_vars = True
                    break
            if any_unbounded_vars:
                continue
            active_parts = r.get_active_partitions()
            assert len(active_parts) == 1
            v, bnds = list(active_parts.items())[0]
            if v in rhs_var_bounds:
                existing_bnds = rhs_var_bounds[v]
                bnds = (max(bnds[0], existing_bnds[0]), min(bnds[1], existing_bnds[1]))
            assert bnds[0] <= bnds[1]
            rhs_var_bounds[v] = bnds
        return integer_var_values, rhs_var_bounds

    @property
    def _elapsed_time(self):
        return time.time() - self._start_time

    @property
    def _remaining_time(self):
        return max(0.0, self.config.time_limit - self._elapsed_time)

    def _perform_obbt(self, vars_to_tighten):
        self._iter += 1
        orig_lbs = list()
        orig_ubs = list()
        for v in vars_to_tighten:
            v_lb, v_ub = v.bounds
            if v_lb is None:
                v_lb = -math.inf
            if v_ub is None:
                v_ub = math.inf
            orig_lbs.append(v_lb)
            orig_ubs.append(v_ub)
        orig_lbs = np.array(orig_lbs)
        orig_ubs = np.array(orig_ubs)
        perform_obbt(self._relaxation, solver=self.mip_solver,
                     varlist=list(vars_to_tighten),
                     objective_bound=self._best_feasible_objective,
                     with_progress_bar=self.config.show_obbt_progress_bar,
                     time_limit=self._remaining_time)
        for r in self._relaxation_objects:
            r.rebuild()
        new_lbs = list()
        new_ubs = list()
        for v in vars_to_tighten:
            v_lb, v_ub = v.bounds
            if v_lb is None:
                v_lb = -math.inf
            if v_ub is None:
                v_ub = math.inf
            new_lbs.append(v_lb)
            new_ubs.append(v_ub)
        new_lbs = np.array(new_lbs)
        new_ubs = np.array(new_ubs)
        lb_diff = new_lbs - orig_lbs
        ub_diff = orig_ubs - new_ubs
        lb_improved = lb_diff > 1e-3
        ub_improved = ub_diff > 1e-3
        lb_improved_indices = lb_improved.nonzero()[0]
        ub_improved_indices = ub_improved.nonzero()[0]
        num_lb_improved = len(lb_improved_indices)
        num_ub_improved = len(ub_improved_indices)
        if num_lb_improved > 0:
            avg_lb_improvement = np.mean(lb_diff[lb_improved_indices])
        else:
            avg_lb_improvement = 0
        if num_ub_improved > 0:
            avg_ub_improvement = np.mean(ub_diff[ub_improved_indices])
        else:
            avg_ub_improvement = 0
        self._log(header=False, num_lb_improved=num_lb_improved,
                  num_ub_improved=num_ub_improved,
                  avg_lb_improvement=avg_lb_improvement,
                  avg_ub_improvement=avg_ub_improvement)

        return num_lb_improved, num_ub_improved, avg_lb_improvement, avg_ub_improvement

    def solve(self, model: _BlockData, timer: HierarchicalTimer = None) -> MultiTreeResults:
        self._re_init()

        self._start_time = time.time()
        if timer is None:
            timer = HierarchicalTimer()
        timer.start("solve")

        self._original_model = model

        self._log(header=True)

        timer.start("construct relaxation")
        self._construct_nlp()
        self._construct_relaxation()
        timer.stop("construct relaxation")

        self._objective = get_objective(self._relaxation)
        self._relaxation_objects = list()
        for r in relaxation_data_objects(
            self._relaxation, descend_into=True, active=True
        ):
            self._relaxation_objects.append(r)

        should_terminate, reason = self._should_terminate()
        if should_terminate:
            return self._get_results(reason)

        self._log(header=False)

        self.mip_solver.set_instance(self._relaxation)
        self._nlp_tightener = appsi.fbbt.IntervalTightener()
        self._nlp_tightener.config.deactivate_satisfied_constraints = True
        self._nlp_tightener.config.feasibility_tol = self.config.feasibility_tolerance
        self._nlp_tightener.set_instance(self._nlp, symbolic_solver_labels=False)

        relaxed_binaries, relaxed_integers = push_integers(self._relaxation)
        self._discrete_vars = list(relaxed_binaries) + list(relaxed_integers)
        oa_results = self._add_oa_cuts(self.config.feasibility_tolerance * 100, 100)
        pop_integers(relaxed_binaries, relaxed_integers)

        should_terminate, reason = self._should_terminate()
        if should_terminate:
            return self._get_results(reason)

        if _is_problem_definitely_convex(self._relaxation):
            oa_results = self._add_oa_cuts(self.config.feasibility_tolerance, 100)
        else:
            oa_results = self._add_oa_cuts(self.config.feasibility_tolerance * 1e3, 3)

        should_terminate, reason = self._should_terminate()
        if should_terminate:
            return self._get_results(reason)

        if oa_results.best_feasible_objective is not None:
            integer_var_values, rhs_var_bounds = self._get_nlp_specs_from_rel()
            nlp_res = self._solve_nlp_with_fixed_vars(
                integer_var_values, rhs_var_bounds
            )

        vars_to_tighten = collect_vars_to_tighten(self._relaxation)
        for obbt_iter in range(self.config.root_obbt_max_iter):
            should_terminate, reason = self._should_terminate()
            if should_terminate:
                return self._get_results(reason)
            relaxed_binaries, relaxed_integers = push_integers(self._relaxation)
            num_lb, num_ub, avg_lb, avg_ub = self._perform_obbt(vars_to_tighten)
            pop_integers(relaxed_binaries, relaxed_integers)
            should_terminate, reason = self._should_terminate()
            if (num_lb + num_ub) < 1 or (avg_lb < 1e-3 and avg_ub < 1e-3):
                break
            if should_terminate:
                return self._get_results(reason)
            self._solve_relaxation()

        while True:
            should_terminate, reason = self._should_terminate()
            if should_terminate:
                break

            rel_res = self._solve_relaxation()

            should_terminate, reason = self._should_terminate()
            if should_terminate:
                break

            if rel_res.best_feasible_objective is not None:
                integer_var_values, rhs_var_bounds = self._get_nlp_specs_from_rel()
                nlp_res = self._solve_nlp_with_fixed_vars(
                    integer_var_values, rhs_var_bounds
                )
                self._oa_cut_helper(self.config.feasibility_tolerance)
                self._partition_helper()
            else:
                self.config.solver_output_logger.warning(
                    f"relaxation did not find a feasible solution: "
                    f"{rel_res.termination_condition}"
                )

        res = self._get_results(reason)

        timer.stop("solve")

        return res



from pyomo.core.base.block import _BlockData, Block
from .custom_block import declare_custom_block
import weakref
import pyomo.environ as pe
from collections.abc import Iterable
from pyomo.common.collections import ComponentSet, ComponentMap
from coramin.utils.coramin_enums import FunctionShape, RelaxationSide
import warnings
import logging
import math
from ._utils import _get_bnds_list, _get_bnds_tuple
import sys
from pyomo.core.expr import taylor_series_expansion
from typing import Sequence, Dict, Tuple, Optional, Union, Mapping, MutableMapping, List
from pyomo.core.base.param import IndexedParam, ScalarParam, _ParamData
from pyomo.core.base.var import _GeneralVarData
from pyomo.core.expr.calculus.diff_with_pyomo import reverse_sd
from pyomo.core.expr.numeric_expr import LinearExpression, ExpressionBase
from pyomo.core.base.constraint import IndexedConstraint, ScalarConstraint, _GeneralConstraintData
from pyomo.contrib.fbbt import interval

pyo = pe
logger = logging.getLogger(__name__)

"""
Base classes for relaxations
"""


class _OACut(object):
    def __init__(self,
                 nonlin_expr,
                 expr_vars: Sequence[_GeneralVarData],
                 coefficients: Sequence[_ParamData],
                 offset: _ParamData):
        self.expr_vars = expr_vars
        self.nonlin_expr = nonlin_expr
        self.coefficients = coefficients
        self.offset = offset
        derivs = reverse_sd(self.nonlin_expr)
        self.derivs = [derivs[i] for i in self.expr_vars]
        self.cut_expr = LinearExpression(constant=self.offset,
                                         linear_coefs=self.coefficients,
                                         linear_vars=self.expr_vars)
        self.current_pt = None

    def update(self,
               var_vals: Sequence[float],
               relaxation_side: RelaxationSide,
               too_small: float,
               too_large: float,
               safety_tol: float) -> Tuple[bool, Optional[_GeneralVarData], Optional[float], Optional[str]]:
        res = (True, None, None, None)
        self.current_pt = var_vals
        orig_values = [i.value for i in self.expr_vars]
        for v, val in zip(self.expr_vars, var_vals):
            v.set_value(val, skip_validation=True)
        try:
            offset_val = pe.value(self.nonlin_expr)
            for ndx, v in enumerate(self.expr_vars):
                der = pe.value(self.derivs[ndx])
                offset_val -= der * v.value
                self.coefficients[ndx]._value = der
            self.offset._value = offset_val
        except (OverflowError, ValueError, ZeroDivisionError) as e:
            res = (False, None, None, str(e))
        finally:
            for v, val in zip(self.expr_vars, orig_values):
                v.set_value(val, skip_validation=True)
        if res[0]:
            res = _check_cut(self.cut_expr, too_small=too_small, too_large=too_large, relaxation_side=relaxation_side,
                             safety_tol=safety_tol)
        return res

    def __repr__(self):
        pt_str = {str(v): p for v, p in zip(self.expr_vars, self.current_pt)}
        pt_str = str(pt_str)
        s = f'OA Cut at {pt_str}'
        return s

    def __str__(self):
        return self.__repr__()


def _check_cut(cut: LinearExpression, too_small, too_large, relaxation_side, safety_tol):
    res = (True, None, None, None)
    for coef_p, v in zip(cut.linear_coefs, cut.linear_vars):
        coef = coef_p.value
        if not math.isfinite(coef) or abs(coef) >= too_large:
            res = (False, v, coef, None)
        elif 0 < abs(coef) <= too_small and v.has_lb() and v.has_ub():
            coef_p._value = 0
            if relaxation_side == RelaxationSide.UNDER:
                cut.constant._value = interval.add(cut.constant.value, cut.constant.value,
                                                   *interval.mul(v.lb, v.ub, coef, coef))[0]
            elif relaxation_side == RelaxationSide.OVER:
                cut.constant._value = interval.add(cut.constant.value, cut.constant.value,
                                                   *interval.mul(v.lb, v.ub, coef, coef))[1]
            else:
                raise ValueError('relaxation_side should be either UNDER or OVER')
    if relaxation_side == RelaxationSide.UNDER:
        cut.constant._value -= safety_tol
    else:
        cut.constant._value += safety_tol
    if not math.isfinite(cut.constant.value) or abs(cut.constant.value) >= too_large:
        res = (False, None, cut.constant.value, None)
    return res


@declare_custom_block(name='BaseRelaxation')
class BaseRelaxationData(_BlockData):
    def __init__(self, component):
        _BlockData.__init__(self, component)
        self._relaxation_side = RelaxationSide.BOTH
        self._use_linear_relaxation = True
        self._large_coef = 1e5
        self._small_coef = 1e-10
        self._needs_rebuilt = True
        self.safety_tol = 1e-10

        self._oa_points: Dict[Tuple[float, ...], _OACut] = dict()
        self._oa_param_indices: MutableMapping[_ParamData, int] = pe.ComponentMap()
        self._current_param_index = 0
        self._oa_params: Optional[IndexedParam] = None
        self._cuts: Optional[IndexedConstraint] = None

        self._saved_oa_points = list()
        self._oa_stack_map = dict()

        self._original_constraint: Optional[ScalarConstraint] = None
        self._nonlinear: Optional[ScalarConstraint] = None

    def set_input(self, relaxation_side=RelaxationSide.BOTH,
                  use_linear_relaxation=True, large_coef=1e5,
                  small_coef=1e-10, safety_tol=1e-10):
        self.relaxation_side = relaxation_side
        self.use_linear_relaxation = use_linear_relaxation
        self._large_coef = large_coef
        self._small_coef = small_coef
        self.safety_tol = safety_tol
        self._needs_rebuilt = True

        self.clear_oa_points()
        self._saved_oa_points = list()
        self._oa_stack_map = dict()

    def get_aux_var(self) -> _GeneralVarData:
        """
        All Coramin relaxations are relaxations of constraints of the form w <=/=/>= f(x). This method returns w

        Returns
        -------
        aux_var: pyomo.core.base.var._GeneralVarData
            The variable representing w in w = f(x) (which is the constraint being relaxed).
        """
        return self._aux_var

    def get_rhs_vars(self) -> Tuple[_GeneralVarData, ...]:
        raise NotImplementedError('This method should be implemented by subclasses')

    def get_rhs_expr(self) -> ExpressionBase:
        raise NotImplementedError('This method should be implemented by subclasses')

    def _get_expr_for_oa(self):
        return self.get_rhs_expr()

    @property
    def small_coef(self):
        return self._small_coef

    @small_coef.setter
    def small_coef(self, val):
        self._small_coef = val

    @property
    def large_coef(self):
        return self._large_coef

    @large_coef.setter
    def large_coef(self, val):
        self._large_coef = val

    @property
    def use_linear_relaxation(self) -> bool:
        """
        If this is True, the relaxation will use a linear relaxation. If False, then a nonlinear relaxation may be used.
        Take x^2 for example, the underestimator can be quadratic.

        Returns
        -------
        bool
        """
        return self._use_linear_relaxation

    @use_linear_relaxation.setter
    def use_linear_relaxation(self, val: bool):
        if not val:
            raise ValueError('Relaxations of type {0} do not support relaxations that are not linear.'.format(type(self)))

    def remove_relaxation(self):
        """
        Remove any auto-created vars/constraints from the relaxation block
        """
        del self._cuts
        self._cuts = None
        del self._original_constraint
        self._original_constraint = None
        del self._nonlinear
        self._nonlinear = None

    def _has_a_convex_side(self):
        if self.has_convex_underestimator() and self.relaxation_side in {RelaxationSide.UNDER, RelaxationSide.BOTH}:
            return True
        if self.has_concave_overestimator() and self.relaxation_side in {RelaxationSide.OVER, RelaxationSide.BOTH}:
            return True
        return False

    def _check_valid_domain_for_relaxation(self) -> bool:
        for v in self.get_rhs_vars():
            lb, ub = _get_bnds_tuple(v)
            if not math.isfinite(lb) or not math.isfinite(ub):
                return False
        return True

    def rebuild(self, build_nonlinear_constraint=False, ensure_oa_at_vertices=True):
        # we have to ensure only one of
        #    - self._cuts
        #    - self._nonlinear
        #    - self._original_constraint
        # is ever not None at one time
        needs_rebuilt = self._needs_rebuilt
        if build_nonlinear_constraint:
            if self._original_constraint is None:
                needs_rebuilt = True
        else:
            if self.use_linear_relaxation:
                if self._nonlinear is not None or self._original_constraint is not None:
                    needs_rebuilt = True
            else:
                if self._cuts is not None or self._original_constraint is not None:
                    needs_rebuilt = True

        if needs_rebuilt:
            self.remove_relaxation()

        self._needs_rebuilt = False

        if build_nonlinear_constraint and self._original_constraint is None:
            del self._original_constraint
            if self.relaxation_side == RelaxationSide.BOTH:
                self._original_constraint = pe.Constraint(expr=self.get_aux_var() == self.get_rhs_expr())
            elif self.relaxation_side == RelaxationSide.UNDER:
                self._original_constraint = pe.Constraint(expr=self.get_aux_var() >= self.get_rhs_expr())
            else:
                self._original_constraint = pe.Constraint(expr=self.get_aux_var() <= self.get_rhs_expr())
        else:
            if self._has_a_convex_side():
                if self.use_linear_relaxation:
                    if self._cuts is None:
                        del self._cuts
                        self._cuts = IndexedConstraint(pe.Any)
                    if self._oa_params is None:
                        del self._oa_params
                        self._oa_params = IndexedParam(pe.Any, mutable=True)
                    self.clean_oa_points(ensure_oa_at_vertices=ensure_oa_at_vertices)
                    self._update_oa_cuts()
                else:
                    if self._nonlinear is None:
                        del self._nonlinear
                        if self.has_convex_underestimator():
                            self._nonlinear = pe.Constraint(expr=self.get_aux_var() >= self._get_expr_for_oa() - self.safety_tol)
                        else:
                            assert self.has_concave_overestimator()
                            self._nonlinear = pe.Constraint(expr=self.get_aux_var() <= self._get_expr_for_oa() + self.safety_tol)

    def vars_with_bounds_in_relaxation(self):
        """
        This method returns a list of variables whose bounds appear in the constraints defining the relaxation.
        Take the McCormick relaxation of a bilinear term (w = x * y) for example. The McCormick relaxation is

        w >= xl * y + x * yl - xl * yl
        w >= xu * y + x * yu - xu * yu
        w <= xu * y + x * yl - xu * yl
        w <= x * yu + xl * y - xl * yu

        where xl and xu are the lower and upper bounds for x, respectively, and yl and yu are the lower and upper
        bounds for y, respectively. Because xl, xu, yl, and yu appear in the constraints, this method would return

        [x, y]

        As another example, take w >= x**2. A linear relaxation of this constraint just involves linear underestimators,
        which do not depend on the bounds of x or w. Therefore, this method would return an empty list.
        """
        raise NotImplementedError('This method should be implemented in the derived class.')

    def get_deviation(self):
        """
        All Coramin relaxations are relaxations of constraints of the form w <=/=/>= f(x). This method returns

        max{f(x) - w, 0} if relaxation_side is RelaxationSide.UNDER
        max{w - f(x), 0} if relaxation_side is RelaxationSide.OVER
        abs(w - f(x)) if relaxation_side is RelaxationSide.BOTH

        Returns
        -------
        float
        """
        dev = self.get_aux_var().value - pe.value(self.get_rhs_expr())
        if self.relaxation_side is RelaxationSide.BOTH:
            dev = abs(dev)
        elif self.relaxation_side is RelaxationSide.UNDER:
            dev = max(-dev, 0)
        else:
            dev = max(dev, 0)
        return dev

    def is_rhs_convex(self):
        """
        All Coramin relaxations are relaxations of constraints of the form w <=/=/>= f(x). This method returns True if f(x)
        is convex and False otherwise.

        Returns
        -------
        bool
        """
        raise NotImplementedError('This method should be implemented in the derived class.')

    def is_rhs_concave(self):
        """
        All Coramin relaxations are relaxations of constraints of the form w <=/=/>= f(x). This method returns True if f(x)
        is concave and False otherwise.

        Returns
        -------
        bool
        """
        raise NotImplementedError('This method should be implemented in the derived class.')

    def has_convex_underestimator(self):
        return self.is_rhs_convex()

    def has_concave_overestimator(self):
        return self.is_rhs_concave()

    @property
    def relaxation_side(self):
        return self._relaxation_side

    @relaxation_side.setter
    def relaxation_side(self, val):
        if val not in RelaxationSide:
            raise ValueError('{0} is not a valid member of RelaxationSide'.format(val))
        if val != self._relaxation_side:
            self._needs_rebuilt = True
        self._relaxation_side = val

    def _get_pprint_string(self):
        if self.relaxation_side == RelaxationSide.BOTH:
            relational_operator_string = '=='
        elif self.relaxation_side == RelaxationSide.UNDER:
            relational_operator_string = '>='
        elif self.relaxation_side == RelaxationSide.OVER:
            relational_operator_string = '<='
        else:
            raise ValueError('Unexpected relaxation side')
        return f'Relaxation for {self.get_aux_var().name} {relational_operator_string} {str(self.get_rhs_expr())}'

    def pprint(self, ostream=None, verbose=False, prefix=""):
        if ostream is None:
            ostream = sys.stdout

        ostream.write('{0}{1}: {2}\n'.format(prefix, self.name, self._get_pprint_string()))

        if verbose:
            super(BaseRelaxationData, self).pprint(ostream=ostream,
                                                   verbose=verbose, prefix=(prefix + '  '))

    def _get_oa_cut(self) -> _OACut:
        rhs_vars = self.get_rhs_vars()
        coef_params = list()
        for v in rhs_vars:
            p = self._oa_params[self._current_param_index]
            self._oa_param_indices[p] = self._current_param_index
            coef_params.append(p)
            self._current_param_index += 1
        offset_param = self._oa_params[self._current_param_index]
        self._oa_param_indices[offset_param] = self._current_param_index
        self._current_param_index += 1
        oa_cut = _OACut(self._get_expr_for_oa(), rhs_vars, coef_params, offset_param)
        return oa_cut

    def _remove_oa_cut(self, oa_cut: _OACut):
        for p in oa_cut.coefficients:
            del self._oa_params[self._oa_param_indices[p]]
            del self._oa_param_indices[p]
        del self._oa_params[self._oa_param_indices[oa_cut.offset]]
        del self._oa_param_indices[oa_cut.offset]
        if oa_cut in self._cuts:  # if the cut did not pass _check_cut, it won't be in self._cuts
            del self._cuts[oa_cut]

    def _log_bad_cut(self, fail_var, fail_coef, err_msg):
        if fail_var is None and fail_coef is None:
            logger.debug(f'Encountered exception when adding OA cut '
                         f'for "{self._get_pprint_string()}"; Error message: {err_msg}')
        elif fail_var is None:
            logger.debug(f'Skipped OA cut for "{self._get_pprint_string()}" due to a '
                         f'large constant value: {fail_coef}')
        else:
            logger.debug(f'Skipped OA cut for "{self._get_pprint_string()}" due to a '
                         f'small or large coefficient for {str(fail_var)}: {fail_coef}')

    def _add_oa_cut(self, pt_tuple: Tuple[float, ...], oa_cut: _OACut) -> Optional[_GeneralConstraintData]:
        if self._nonlinear is not None or self._original_constraint is not None:
            raise ValueError('Can only add an OA cut when using a linear relaxation')
        if self.has_convex_underestimator():
            rel_side = RelaxationSide.UNDER
        else:
            assert self.has_concave_overestimator()
            rel_side = RelaxationSide.OVER
        cut_info = oa_cut.update(var_vals=pt_tuple, relaxation_side=rel_side,
                                 too_small=self.small_coef, too_large=self.large_coef,
                                 safety_tol=self.safety_tol)
        success, fail_var, fail_coef, err_msg = cut_info
        if not success:
            self._log_bad_cut(fail_var, fail_coef, err_msg)
            if oa_cut in self._cuts:
                del self._cuts[oa_cut]
        else:
            if oa_cut not in self._cuts:
                if self.has_convex_underestimator():
                    self._cuts[oa_cut] = self.get_aux_var() >= oa_cut.cut_expr
                else:
                    self._cuts[oa_cut] = self.get_aux_var() <= oa_cut.cut_expr
                return self._cuts[oa_cut]
        return None

    def _update_oa_cuts(self):
        for pt_tuple, oa_cut in self._oa_points.items():
            self._add_oa_cut(pt_tuple, oa_cut)

        # remove any cuts that may have been added with add_cut(keep_cut=False)
        all_oa_cuts = set(self._oa_points.values())
        for oa_cut in list(self._cuts):
            if oa_cut not in all_oa_cuts:
                self._remove_oa_cut(oa_cut)

    def _add_oa_point(self, pt_tuple: Tuple[float, ...]):
        if pt_tuple not in self._oa_points:
            self._oa_points[pt_tuple] = self._get_oa_cut()

    def add_oa_point(self, var_values: Optional[Union[Tuple[float, ...], Mapping[_GeneralVarData, float]]] = None):
        """
        Add a point at which an outer-approximation cut for a convex constraint should be added. This does not
        rebuild the relaxation. You must call rebuild() for the constraint to get added.

        Parameters
        ----------
        var_values: Optional[Union[Tuple[float, ...], Mapping[_GeneralVarData, float]]]
        """
        if self._has_a_convex_side():
            if var_values is None:
                var_values = tuple(v.value for v in self.get_rhs_vars())
            elif type(var_values) is tuple:
                pass
            else:
                var_values = tuple(var_values[v] for v in self.get_rhs_vars())
            self._add_oa_point(var_values)

    def push_oa_points(self, key=None):
        """
        Save the current list of OA points for later use through pop_oa_points().
        """
        to_save = [i for i in self._oa_points.keys()]
        if key is not None:
            self._oa_stack_map[key] = to_save
        else:
            self._saved_oa_points.append(to_save)

    def clear_oa_points(self):
        """
        Delete any existing OA points.
        """
        self._oa_points = dict()
        self._oa_param_indices = pe.ComponentMap()
        self._current_param_index = 0
        if self._oa_params is not None:
            del self._oa_params
            self._oa_params = pe.Param(pe.Any, mutable=True)
        if self._cuts is not None:
            del self._cuts
            self._cuts = pe.Constraint(pe.Any)

    def pop_oa_points(self, key=None):
        """
        Use the most recently saved list of OA points
        """
        self.clear_oa_points()
        if key is None:
            list_of_points = self._saved_oa_points.pop(-1)
        else:
            list_of_points = self._oa_stack_map.pop(key)
        for pt_tuple in list_of_points:
            self._add_oa_point(pt_tuple)

    def add_cut(self, keep_cut=True, check_violation=True, feasibility_tol=1e-8) -> Optional[_GeneralConstraintData]:
        """
        This function will add a linear cut to the relaxation. Cuts are only generated for the convex side of the
        constraint (if the constraint has a convex side). For example, if the relaxation is a PWXSquaredRelaxationData
        for y = x**2, the add_cut will add an underestimator at x.value (but only if y.value < x.value**2). If
        relaxation is a PWXSquaredRelaxationData for y < x**2, then no cut will be added. If relaxation is is a
        PWMcCormickRelaxationData, then no cut will be added.

        Parameters
        ----------
        keep_cut: bool
            If keep_cut is True, then add_oa_point will also be called. Be careful if the relaxation object is relaxing
            the nonconvex side of the constraint. Thus, the cut will be reconstructed when rebuild is called. If
            keep_cut is False, then the cut will be discarded when rebuild is called.
        check_violation: bool
            If True, then a cut is only added if the cut generated would cut off the current point (current values
            of the variables) by more than feasibility_tol.
        feasibility_tol: float
            Only used if check_violation is True

        Returns
        -------
        new_con: pyomo.core.base.constraint._GeneralConstraintData
        """
        rhs_vars = self.get_rhs_vars()
        var_vals = tuple(v.value for v in rhs_vars)

        if var_vals in self._oa_points:
            return None

        new_con = None
        if self._has_a_convex_side():
            if check_violation:
                needs_cut = False
                try:
                    rhs_val = pe.value(self._get_expr_for_oa())
                except (OverflowError, ZeroDivisionError, ValueError):
                    rhs_val = None
                if rhs_val is not None:
                    if self.has_convex_underestimator():
                        viol = rhs_val - self.get_aux_var().value
                    else:
                        viol = self.get_aux_var().value - rhs_val
                    if viol > feasibility_tol:
                        needs_cut = True
            else:
                needs_cut = True
            if needs_cut:
                oa_cut = self._get_oa_cut()
                new_con = self._add_oa_cut(pt_tuple=var_vals, oa_cut=oa_cut)
                if keep_cut:
                    self._oa_points[var_vals] = oa_cut

        return new_con

    def clean_oa_points(self, ensure_oa_at_vertices=True):
        if not self._has_a_convex_side():
            return

        rhs_vars = self.get_rhs_vars()
        bnds_list: List[Tuple[float, float]] = list()
        for v in rhs_vars:
            bnds_list.append(_get_bnds_tuple(v))

        for pt_tuple, oa_cut in list(self._oa_points.items()):
            new_pt_list = list()
            for (v_lb, v_ub), pt in zip(bnds_list, pt_tuple):
                if pt < v_lb:
                    new_pt_list.append(v_lb)
                elif pt > v_ub:
                    new_pt_list.append(v_ub)
                else:
                    new_pt_list.append(pt)
            new_pt_tuple = tuple(new_pt_list)
            del self._oa_points[pt_tuple]
            if new_pt_tuple in self._oa_points:
                self._remove_oa_cut(oa_cut)
            else:
                self._oa_points[new_pt_tuple] = oa_cut
        if ensure_oa_at_vertices:
            lb_list = list()
            ub_list = list()
            for lb, ub in bnds_list:
                if math.isfinite(lb) and math.isfinite(ub):
                    lb_list.append(lb)
                    ub_list.append(ub)
                elif math.isfinite(lb):
                    lb_list.append(lb)
                    ub_list.append(max(lb + 1, 1))
                elif math.isfinite(ub):
                    lb_list.append(min(ub - 1, -1))
                    ub_list.append(ub)
                else:
                    lb_list.append(-1)
                    ub_list.append(1)
            lb_tuple = tuple(lb_list)
            ub_tuple = tuple(ub_list)
            if lb_tuple not in self._oa_points:
                if len(self._oa_points) <= 1:
                    self._add_oa_point(lb_tuple)
                else:  # move the smallest point to lb_tuple
                    min_pt = min(self._oa_points.keys())
                    min_oa_cut = self._oa_points[min_pt]
                    del self._oa_points[min_pt]
                    self._oa_points[lb_tuple] = min_oa_cut
            if ub_tuple not in self._oa_points:
                if len(self._oa_points) <= 1:
                    self._add_oa_point(ub_tuple)
                else: # move the largest point to ub_tuple
                    max_pt = max(self._oa_points.keys())
                    max_oa_cut = self._oa_points[max_pt]
                    del self._oa_points[max_pt]
                    self._oa_points[ub_tuple] = max_oa_cut


@declare_custom_block(name='BasePWRelaxation')
class BasePWRelaxationData(BaseRelaxationData):
    def __init__(self, component):
        BaseRelaxationData.__init__(self, component)

        self._partitions = ComponentMap()  # ComponentMap: var: list of float
        self._saved_partitions = list()  # list of CompnentMap

    def rebuild(self, build_nonlinear_constraint=False, ensure_oa_at_vertices=True):
        """
        Remove any auto-created vars/constraints from the relaxation block and recreate it
        """
        super(BasePWRelaxationData, self).rebuild(build_nonlinear_constraint=build_nonlinear_constraint,
                                                  ensure_oa_at_vertices=ensure_oa_at_vertices)
        self.clean_partitions()

    def set_input(self, relaxation_side=RelaxationSide.BOTH, use_linear_relaxation=True, large_coef=1e5,
                  small_coef=1e-10, safety_tol=1e-10):
        super(BasePWRelaxationData, self).set_input(relaxation_side=relaxation_side,
                                                    use_linear_relaxation=use_linear_relaxation, large_coef=large_coef,
                                                    small_coef=small_coef, safety_tol=safety_tol)
        self._partitions = ComponentMap()
        self._saved_partitions = list()

    def add_partition_point(self):
        """
        Add a point to the current partitioning. This does not rebuild the relaxation. You must call rebuild()
        to rebuild the relaxation.
        """
        raise NotImplementedError('This method should be implemented in the derived class.')

    def _add_partition_point(self, var, value=None):
        if value is None:
            value = pe.value(var)
        # if the point is outside the variable's bounds, then it will simply get removed when clean_partitions
        # gets called.
        self._partitions[var].append(value)

    def push_partitions(self):
        """
        Save the current partitioning for later use through pop_partitions().
        """
        self._saved_partitions.append(pe.ComponentMap((k, list(v)) for k, v in self._partitions.items()))

    def clear_partitions(self):
        """
        Delete any existing partitioning scheme.
        """
        tmp = ComponentMap()
        for var, pts in self._partitions.items():
            tmp[var] = [pe.value(var.lb), pe.value(var.ub)]
        self._partitions = tmp

    def pop_partitions(self):
        """
        Use the most recently saved partitioning.
        """
        self._partitions = self._saved_partitions.pop(-1)

    def clean_partitions(self):
        # discard any points in the partitioning that are not within the variable bounds
        for var, pts in list(self._partitions.items()):
            pts = list(set(pts))
            pts.sort()
            self._partitions[var] = pts

        for var, pts in self._partitions.items():
            lb, ub = tuple(_get_bnds_list(var))

            new_pts = list()
            new_pts.append(lb)
            for val in pts[1:-1]:
                if lb < val < ub:
                    new_pts.append(val)
            new_pts.append(ub)
            self._partitions[var] = new_pts

    def get_active_partitions(self):
        ans = ComponentMap()
        for var, pts in self._partitions.items():
            val = pyo.value(var)
            lower = var.lb
            upper = var.ub
            for p in pts:
                if val >= p and p > lower:
                    lower = p
                if val <= p and p < upper:
                    upper = p
            ans[var] = lower, upper
        return ans


class ComponentWeakRef(object):
    """
    This object is used to reference components from a block that are not owned by that block.
    """
    # ToDo: Example in the documentation
    def __init__(self, comp):
        self.compref = None
        self.set_component(comp)

    def get_component(self):
        if self.compref is None:
            return None
        return self.compref()

    def set_component(self, comp):
        self.compref = None
        if comp is not None:
            self.compref = weakref.ref(comp)

    def __setstate__(self, state):
        self.set_component(state['compref'])

    def __getstate__(self):
        return {'compref': self.get_component()}



import pyomo.environ as pyo
from coramin.utils.coramin_enums import RelaxationSide, FunctionShape
from .relaxations_base import BasePWRelaxationData, ComponentWeakRef, _check_cut
from .custom_block import declare_custom_block
import numpy as np
import math
import scipy.optimize
from ._utils import check_var_pts, _get_bnds_list, _get_bnds_tuple
from pyomo.core.base.param import ScalarParam, IndexedParam
from pyomo.core.base.constraint import ScalarConstraint, IndexedConstraint
from pyomo.core.expr.numeric_expr import LinearExpression
import logging
from typing import Optional, Union, Sequence
from pyomo.core.expr.calculus.diff_with_pyomo import reverse_sd
logger = logging.getLogger(__name__)
pe = pyo


def _sin_overestimator_fn(x, LB):
    return np.sin(x) + np.cos(x) * (LB - x) - np.sin(LB)


def _sin_underestimator_fn(x, UB):
    return np.sin(x) + np.cos(-x) * (UB - x) - np.sin(UB)


def _compute_sine_overestimator_tangent_point(vlb):
    assert vlb < 0
    tangent_point, res = scipy.optimize.bisect(f=_sin_overestimator_fn, a=0, b=math.pi / 2, args=(vlb,),
                                               full_output=True, disp=False)
    if res.converged:
        tangent_point = float(tangent_point)
        slope = float(np.cos(tangent_point))
        intercept = float(np.sin(vlb) - slope * vlb)
        return tangent_point, slope, intercept
    else:
        raise RuntimeError('Unable to build relaxation for sin(x)\nBisect info: ' + str(res))


def _compute_sine_underestimator_tangent_point(vub):
    assert vub > 0
    tangent_point, res = scipy.optimize.bisect(f=_sin_underestimator_fn, a=-math.pi / 2, b=0, args=(vub,),
                                               full_output=True, disp=False)
    if res.converged:
        tangent_point = float(tangent_point)
        slope = float(np.cos(-tangent_point))
        intercept = float(np.sin(vub) - slope * vub)
        return tangent_point, slope, intercept
    else:
        raise RuntimeError('Unable to build relaxation for sin(x)\nBisect info: ' + str(res))


def _atan_overestimator_fn(x, LB):
    return (1 + x**2) * (np.arctan(x) - np.arctan(LB)) - x + LB


def _atan_underestimator_fn(x, UB):
    return (1 + x**2) * (np.arctan(x) - np.arctan(UB)) - x + UB


def _compute_arctan_overestimator_tangent_point(vlb):
    assert vlb < 0
    tangent_point, res = scipy.optimize.bisect(f=_atan_overestimator_fn, a=0, b=abs(vlb), args=(vlb,),
                                               full_output=True, disp=False)
    if res.converged:
        tangent_point = float(tangent_point)
        slope = 1/(1 + tangent_point**2)
        intercept = float(np.arctan(vlb) - slope * vlb)
        return tangent_point, slope, intercept
    else:
        raise RuntimeError('Unable to build relaxation for arctan(x)\nBisect info: ' + str(res))


def _compute_arctan_underestimator_tangent_point(vub):
    assert vub > 0
    tangent_point, res = scipy.optimize.bisect(f=_atan_underestimator_fn, a=-vub, b=0, args=(vub,),
                                               full_output=True, disp=False)
    if res.converged:
        tangent_point = float(tangent_point)
        slope = 1/(1 + tangent_point**2)
        intercept = float(np.arctan(vub) - slope * vub)
        return tangent_point, slope, intercept
    else:
        raise RuntimeError('Unable to build relaxation for arctan(x)\nBisect info: ' + str(res))


class _FxExpr(object):
    def __init__(self, expr, x):
        self._expr = expr
        self._x = x
        self._deriv = reverse_sd(expr)[x]

    def eval(self, _xval):
        _xval = pyo.value(_xval)
        orig_xval = self._x.value
        self._x.value = _xval
        res = pyo.value(self._expr)
        self._x.set_value(orig_xval, skip_validation=True)
        return res

    def deriv(self, _xval):
        _xval = pyo.value(_xval)
        orig_xval = self._x.value
        self._x.value = _xval
        res = pyo.value(self._deriv)
        self._x.set_value(orig_xval, skip_validation=True)
        return res

    def __call__(self, _xval):
        return self.eval(_xval)


def _func_wrapper(obj):
    def _func(m, val):
        return obj(val)
    return _func


def _pw_univariate_relaxation(b, x, w, x_pts, f_x_expr, pw_repn='INC', shape=FunctionShape.UNKNOWN,
                              relaxation_side=RelaxationSide.BOTH, large_eval_tol=math.inf,
                              safety_tol=0):
    """
    This function creates piecewise envelopes to relax "w=f(x)" where f(x) is univariate and either convex over the
    entire domain of x or concave over the entire domain of x.

    Parameters
    ----------
    b: pyo.Block
    x: pyo.Var
        The "x" variable in f(x)
    w: pyo.Var
        The "w" variable that is replacing f(x)
    x_pts: Sequence[float]
        A list of floating point numbers to define the points over which the piecewise representation will generated.
        This list must be ordered, and it is expected that the first point (x_pts[0]) is equal to x.lb and the last
        point (x_pts[-1]) is equal to x.ub
    f_x_expr: pyomo expression
        An expression for f(x)
    pw_repn: str
        This must be one of the valid strings for the peicewise representation to use (directly from the Piecewise
        component). Use help(Piecewise) to learn more.
    shape: FunctionShape
        Specify the shape of the function. Valid values are minlp.FunctionShape.CONVEX or minlp.FunctionShape.CONCAVE
    relaxation_side: RelaxationSide
        Provide the desired side for the relaxation (OVER, UNDER, or BOTH)
    large_eval_tol: float
        To avoid numerical problems, if f_x_expr or its derivative evaluates to a value larger than large_eval_tol, 
        at a point in x_pts, then that point is skipped.
    """
    assert shape in {FunctionShape.CONCAVE, FunctionShape.CONVEX}
    assert relaxation_side in {RelaxationSide.UNDER, RelaxationSide.OVER}
    if relaxation_side == RelaxationSide.UNDER:
        assert shape == FunctionShape.CONCAVE
    else:
        assert shape == FunctionShape.CONVEX

    _eval = _FxExpr(expr=f_x_expr, x=x)
    xlb = x_pts[0]
    xub = x_pts[-1]

    check_var_pts(x, x_pts)

    if x.is_fixed():
        b.x_fixed_con = pyo.Constraint(expr=w == _eval(x.value))
    elif xlb == xub:
        b.x_fixed_con = pyo.Constraint(expr=w == _eval(x.lb))
    else:
        # Do the non-convex piecewise portion if shape=CONCAVE and relaxation_side=Under/BOTH
        # or if shape=CONVEX and relaxation_side=Over/BOTH
        pw_constr_type = None
        if shape == FunctionShape.CONVEX and relaxation_side in {RelaxationSide.OVER, RelaxationSide.BOTH}:
            pw_constr_type = 'UB'
            _eval = _FxExpr(expr=f_x_expr + safety_tol, x=x)
        if shape == FunctionShape.CONCAVE and relaxation_side in {RelaxationSide.UNDER, RelaxationSide.BOTH}:
            pw_constr_type = 'LB'
            _eval = _FxExpr(expr=f_x_expr - safety_tol, x=x)

        if pw_constr_type is not None:
            # Build the piecewise side of the envelope
            if x_pts[0] > -math.inf and x_pts[-1] < math.inf:
                tmp_pts = list()
                for _pt in x_pts:
                    try:
                        f = _eval(_pt)
                        if abs(f) >= large_eval_tol:
                            logger.warning(f'Skipping pt {_pt} for var {str(x)} because |{str(f_x_expr)}| '
                                           f'evaluated at {_pt} is larger than {large_eval_tol}')
                            continue
                        tmp_pts.append(_pt)
                    except (ZeroDivisionError, ValueError, OverflowError):
                        pass
                if len(tmp_pts) >= 2 and tmp_pts[0] == x_pts[0] and tmp_pts[-1] == x_pts[-1]:
                    b.pw_linear_under_over = pyo.Piecewise(w, x,
                                                           pw_pts=tmp_pts,
                                                           pw_repn=pw_repn,
                                                           pw_constr_type=pw_constr_type,
                                                           f_rule=_func_wrapper(_eval)
                                                           )


def pw_sin_relaxation(b, x, w, x_pts, relaxation_side=RelaxationSide.BOTH, safety_tol=1e-10):
    """
    This function creates piecewise relaxations to relax "w=sin(x)" for -pi/2 <= x <= pi/2.

    Parameters
    ----------
    b: pyo.Block
    x: pyomo.core.base.var.SimpleVar or pyomo.core.base.var._GeneralVarData
        The "x" variable in sin(x). The lower bound on x must greater than or equal to
        -pi/2 and the upper bound on x must be less than or equal to pi/2.
    w: pyomo.core.base.var.SimpleVar or pyomo.core.base.var._GeneralVarData
        The auxillary variable replacing sin(x)
    x_pts: Sequence[float]
        A list of floating point numbers to define the points over which the piecewise
        representation will be generated. This list must be ordered, and it is expected
        that the first point (x_pts[0]) is equal to x.lb and the last point (x_pts[-1])
        is equal to x.ub
    relaxation_side: minlp.RelaxationSide
        Provide the desired side for the relaxation (OVER, UNDER, or BOTH)
    safety_tol: float
        amount to lift the overestimator or drop the underestimator. This is used to ensure none of the feasible
        region is cut off by error in computing the over and under estimators.
    """
    check_var_pts(x, x_pts)
    expr = pyo.sin(x)

    xlb = x_pts[0]
    xub = x_pts[-1]

    if x.is_fixed() or xlb == xub:
        b.x_fixed_con = pyo.Constraint(expr=w == (pyo.value(expr)))
        return

    if xlb < -np.pi / 2.0:
        return

    if xub > np.pi / 2.0:
        return

    OE_tangent_x, OE_tangent_slope, OE_tangent_intercept = _compute_sine_overestimator_tangent_point(xlb)
    UE_tangent_x, UE_tangent_slope, UE_tangent_intercept = _compute_sine_underestimator_tangent_point(xub)
    non_piecewise_overestimators_pts = []
    non_piecewise_underestimator_pts = []

    if relaxation_side == RelaxationSide.OVER:
        if OE_tangent_x < xub:
            new_x_pts = [i for i in x_pts if i < OE_tangent_x]
            new_x_pts.append(xub)
            non_piecewise_overestimators_pts = [OE_tangent_x]
            non_piecewise_overestimators_pts.extend(i for i in x_pts if i > OE_tangent_x)
            x_pts = new_x_pts
    elif relaxation_side == RelaxationSide.UNDER:
        if UE_tangent_x > xlb:
            new_x_pts = [xlb]
            new_x_pts.extend(i for i in x_pts if i > UE_tangent_x)
            non_piecewise_underestimator_pts = [i for i in x_pts if i < UE_tangent_x]
            non_piecewise_underestimator_pts.append(UE_tangent_x)
            x_pts = new_x_pts

    b.non_piecewise_overestimators = pyo.ConstraintList()
    b.non_piecewise_underestimators = pyo.ConstraintList()
    for pt in non_piecewise_overestimators_pts:
        b.non_piecewise_overestimators.add(w <= math.sin(pt) + safety_tol + (x - pt) * math.cos(pt))
    for pt in non_piecewise_underestimator_pts:
        b.non_piecewise_underestimators.add(w >= math.sin(pt) - safety_tol + (x - pt) * math.cos(pt))

    intervals = []
    for i in range(len(x_pts)-1):
        intervals.append((x_pts[i], x_pts[i+1]))

    b.interval_set = pyo.Set(initialize=range(len(intervals)), ordered=True)
    b.x = pyo.Var(b.interval_set)
    b.w = pyo.Var(b.interval_set)
    if len(intervals) == 1:
        b.lam = pyo.Param(b.interval_set, mutable=True)
        b.lam[0].value = 1.0
    else:
        b.lam = pyo.Var(b.interval_set, within=pyo.Binary)
    b.x_lb = pyo.ConstraintList()
    b.x_ub = pyo.ConstraintList()
    b.x_sum = pyo.Constraint(expr=x == sum(b.x[i] for i in b.interval_set))
    b.w_sum = pyo.Constraint(expr=w == sum(b.w[i] for i in b.interval_set))
    b.lam_sum = pyo.Constraint(expr=sum(b.lam[i] for i in b.interval_set) == 1)
    b.overestimators = pyo.ConstraintList()
    b.underestimators = pyo.ConstraintList()

    for i, tup in enumerate(intervals):
        x0 = tup[0]
        x1 = tup[1]

        b.x_lb.add(x0 * b.lam[i] <= b.x[i])
        b.x_ub.add(b.x[i] <= x1 * b.lam[i])

        # Overestimators
        if relaxation_side in {RelaxationSide.OVER, RelaxationSide.BOTH}:
            if x0 < 0 and x1 <= 0:
                slope = (math.sin(x1) - math.sin(x0)) / (x1 - x0)
                intercept = math.sin(x0) - slope * x0
                b.overestimators.add(b.w[i] <= slope * b.x[i] + (intercept + safety_tol) * b.lam[i])
            elif (x0 < 0) and (x1 > 0):
                tangent_x, tangent_slope, tangent_intercept = _compute_sine_overestimator_tangent_point(x0)
                if tangent_x <= x1:
                    b.overestimators.add(b.w[i] <= tangent_slope * b.x[i] + (tangent_intercept + safety_tol) * b.lam[i])
                    b.overestimators.add(b.w[i] <= math.cos(x1) * b.x[i] +
                                         (math.sin(x1) - x1 * math.cos(x1) + safety_tol) * b.lam[i])
                else:
                    slope = (math.sin(x1) - math.sin(x0)) / (x1 - x0)
                    intercept = math.sin(x0) - slope * x0
                    b.overestimators.add(b.w[i] <= slope * b.x[i] + (intercept + safety_tol) * b.lam[i])
            else:
                b.overestimators.add(b.w[i] <= math.cos(x0)*b.x[i] +
                                     (math.sin(x0) - x0*math.cos(x0) + safety_tol)*b.lam[i])
                b.overestimators.add(b.w[i] <= math.cos(x1)*b.x[i] +
                                     (math.sin(x1) - x1*math.cos(x1) + safety_tol)*b.lam[i])

        # Underestimators
        if relaxation_side in {RelaxationSide.UNDER, RelaxationSide.BOTH}:
            if x0 >= 0 and x1 > 0:
                slope = (math.sin(x1) - math.sin(x0)) / (x1 - x0)
                intercept = math.sin(x0) - slope * x0
                b.underestimators.add(b.w[i] >= slope * b.x[i] + (intercept - safety_tol) * b.lam[i])
            elif (x1 > 0) and (x0 < 0):
                tangent_x, tangent_slope, tangent_intercept = _compute_sine_underestimator_tangent_point(x1)
                if tangent_x >= x0:
                    b.underestimators.add(b.w[i] >= tangent_slope*b.x[i] + (tangent_intercept - safety_tol)*b.lam[i])
                    b.underestimators.add(b.w[i] >= math.cos(x0)*b.x[i] +
                                          (math.sin(x0) - x0 * math.cos(x0) - safety_tol)*b.lam[i])
                else:
                    slope = (math.sin(x1) - math.sin(x0)) / (x1 - x0)
                    intercept = math.sin(x0) - slope * x0
                    b.underestimators.add(b.w[i] >= slope * b.x[i] + (intercept - safety_tol) * b.lam[i])
            else:
                b.underestimators.add(b.w[i] >= math.cos(x0)*b.x[i] +
                                      (math.sin(x0) - x0 * math.cos(x0) - safety_tol)*b.lam[i])
                b.underestimators.add(b.w[i] >= math.cos(x1)*b.x[i] +
                                      (math.sin(x1) - x1 * math.cos(x1) - safety_tol)*b.lam[i])

    return x_pts


def pw_arctan_relaxation(b, x, w, x_pts, relaxation_side=RelaxationSide.BOTH, safety_tol=1e-10):
    """
    This function creates piecewise relaxations to relax "w=sin(x)" for -pi/2 <= x <= pi/2.

    Parameters
    ----------
    b: pyo.Block
    x: pyomo.core.base.var.SimpleVar or pyomo.core.base.var._GeneralVarData
        The "x" variable in sin(x). The lower bound on x must greater than or equal to
        -pi/2 and the upper bound on x must be less than or equal to pi/2.
    w: pyomo.core.base.var.SimpleVar or pyomo.core.base.var._GeneralVarData
        The auxillary variable replacing sin(x)
    x_pts: Sequence[float]
        A list of floating point numbers to define the points over which the piecewise
        representation will be generated. This list must be ordered, and it is expected
        that the first point (x_pts[0]) is equal to x.lb and the last point (x_pts[-1])
        is equal to x.ub
    relaxation_side: minlp.RelaxationSide
        Provide the desired side for the relaxation (OVER, UNDER, or BOTH)
    safety_tol: float
        amount to lift the overestimator or drop the underestimator. This is used to ensure none of the feasible
        region is cut off by error in computing the over and under estimators.
    """
    check_var_pts(x, x_pts)
    expr = pyo.atan(x)
    _eval = _FxExpr(expr, x)

    xlb = x_pts[0]
    xub = x_pts[-1]

    if x.is_fixed() or xlb == xub:
        b.x_fixed_con = pyo.Constraint(expr=w == pyo.value(expr))
        return

    if xlb == -math.inf or xub == math.inf:
        return

    OE_tangent_x, OE_tangent_slope, OE_tangent_intercept = _compute_arctan_overestimator_tangent_point(xlb)
    UE_tangent_x, UE_tangent_slope, UE_tangent_intercept = _compute_arctan_underestimator_tangent_point(xub)
    non_piecewise_overestimators_pts = []
    non_piecewise_underestimator_pts = []

    if relaxation_side == RelaxationSide.OVER:
        if OE_tangent_x < xub:
            new_x_pts = [i for i in x_pts if i < OE_tangent_x]
            new_x_pts.append(xub)
            non_piecewise_overestimators_pts = [OE_tangent_x]
            non_piecewise_overestimators_pts.extend(i for i in x_pts if i > OE_tangent_x)
            x_pts = new_x_pts
    elif relaxation_side == RelaxationSide.UNDER:
        if UE_tangent_x > xlb:
            new_x_pts = [xlb]
            new_x_pts.extend(i for i in x_pts if i > UE_tangent_x)
            non_piecewise_underestimator_pts = [i for i in x_pts if i < UE_tangent_x]
            non_piecewise_underestimator_pts.append(UE_tangent_x)
            x_pts = new_x_pts

    b.non_piecewise_overestimators = pyo.ConstraintList()
    b.non_piecewise_underestimators = pyo.ConstraintList()
    for pt in non_piecewise_overestimators_pts:
        b.non_piecewise_overestimators.add(w <= math.atan(pt) + safety_tol + (x - pt) * _eval.deriv(pt))
    for pt in non_piecewise_underestimator_pts:
        b.non_piecewise_underestimators.add(w >= math.atan(pt) - safety_tol + (x - pt) * _eval.deriv(pt))

    intervals = []
    for i in range(len(x_pts)-1):
        intervals.append((x_pts[i], x_pts[i+1]))

    b.interval_set = pyo.Set(initialize=range(len(intervals)))
    b.x = pyo.Var(b.interval_set)
    b.w = pyo.Var(b.interval_set)
    if len(intervals) == 1:
        b.lam = pyo.Param(b.interval_set, mutable=True)
        b.lam[0].value = 1.0
    else:
        b.lam = pyo.Var(b.interval_set, within=pyo.Binary)
    b.x_lb = pyo.ConstraintList()
    b.x_ub = pyo.ConstraintList()
    b.x_sum = pyo.Constraint(expr=x == sum(b.x[i] for i in b.interval_set))
    b.w_sum = pyo.Constraint(expr=w == sum(b.w[i] for i in b.interval_set))
    b.lam_sum = pyo.Constraint(expr=sum(b.lam[i] for i in b.interval_set) == 1)
    b.overestimators = pyo.ConstraintList()
    b.underestimators = pyo.ConstraintList()

    for i, tup in enumerate(intervals):
        x0 = tup[0]
        x1 = tup[1]

        b.x_lb.add(x0 * b.lam[i] <= b.x[i])
        b.x_ub.add(b.x[i] <= x1 * b.lam[i])

        # Overestimators
        if relaxation_side in {RelaxationSide.OVER, RelaxationSide.BOTH}:
            if x0 < 0 and x1 <= 0:
                slope = (math.atan(x1) - math.atan(x0)) / (x1 - x0)
                intercept = math.atan(x0) - slope * x0
                b.overestimators.add(b.w[i] <= slope * b.x[i] + (intercept + safety_tol) * b.lam[i])
            elif (x0 < 0) and (x1 > 0):
                tangent_x, tangent_slope, tangent_intercept = _compute_arctan_overestimator_tangent_point(x0)
                if tangent_x <= x1:
                    b.overestimators.add(b.w[i] <= tangent_slope * b.x[i] + (tangent_intercept + safety_tol) * b.lam[i])
                    b.overestimators.add(b.w[i] <= _eval.deriv(x1)*b.x[i] +
                                         (math.atan(x1) - x1*_eval.deriv(x1) + safety_tol)*b.lam[i])
                else:
                    slope = (math.atan(x1) - math.atan(x0)) / (x1 - x0)
                    intercept = math.atan(x0) - slope * x0
                    b.overestimators.add(b.w[i] <= slope * b.x[i] + (intercept + safety_tol) * b.lam[i])
            else:
                b.overestimators.add(b.w[i] <= _eval.deriv(x0)*b.x[i] +
                                     (math.atan(x0) - x0*_eval.deriv(x0) + safety_tol)*b.lam[i])
                b.overestimators.add(b.w[i] <= _eval.deriv(x1)*b.x[i] +
                                     (math.atan(x1) - x1*_eval.deriv(x1) + safety_tol)*b.lam[i])

        # Underestimators
        if relaxation_side in {RelaxationSide.UNDER, RelaxationSide.BOTH}:
            if x0 >= 0 and x1 > 0:
                slope = (math.atan(x1) - math.atan(x0)) / (x1 - x0)
                intercept = math.atan(x0) - slope * x0
                b.underestimators.add(b.w[i] >= slope * b.x[i] + (intercept - safety_tol) * b.lam[i])
            elif (x1 > 0) and (x0 < 0):
                tangent_x, tangent_slope, tangent_intercept = _compute_arctan_underestimator_tangent_point(x1)
                if tangent_x >= x0:
                    b.underestimators.add(b.w[i] >= tangent_slope*b.x[i] + (tangent_intercept - safety_tol)*b.lam[i])
                    b.underestimators.add(b.w[i] >= _eval.deriv(x0)*b.x[i] +
                                          (math.atan(x0) - x0*_eval.deriv(x0) - safety_tol)*b.lam[i])
                else:
                    slope = (math.atan(x1) - math.atan(x0)) / (x1 - x0)
                    intercept = math.atan(x0) - slope * x0
                    b.underestimators.add(b.w[i] >= slope * b.x[i] + (intercept - safety_tol) * b.lam[i])
            else:
                b.underestimators.add(b.w[i] >= _eval.deriv(x0)*b.x[i] +
                                      (math.atan(x0) - x0*_eval.deriv(x0) - safety_tol)*b.lam[i])
                b.underestimators.add(b.w[i] >= _eval.deriv(x1)*b.x[i] +
                                      (math.atan(x1) - x1*_eval.deriv(x1) - safety_tol)*b.lam[i])

    return x_pts


@declare_custom_block(name='PWUnivariateRelaxation')
class PWUnivariateRelaxationData(BasePWRelaxationData):
    """
    A helper class for building and modifying piecewise relaxations of aux_var = f(x) where f(x) is either convex
    or concave.
    """

    def __init__(self, component):
        super().__init__(component)
        self._xref = ComponentWeakRef(None)
        self._aux_var_ref = ComponentWeakRef(None)
        self._pw_repn = 'INC'
        self._function_shape = FunctionShape.UNKNOWN
        self._f_x_expr = None
        self._secant: Optional[Union[ScalarConstraint, IndexedConstraint]] = None
        self._secant_expr: Optional[LinearExpression] = None
        self._secant_slope: Optional[Union[ScalarParam, IndexedParam]] = None
        self._secant_intercept: Optional[Union[ScalarParam, IndexedParam]] = None
        self._pw_secant = None

    @property
    def _x(self):
        return self._xref.get_component()

    @property
    def _aux_var(self):
        return self._aux_var_ref.get_component()

    def get_rhs_vars(self):
        return self._x,

    def get_rhs_expr(self):
        return self._f_x_expr

    def vars_with_bounds_in_relaxation(self):
        res = list()
        if self.relaxation_side == RelaxationSide.BOTH:
            res.append(self._x)
        elif self.relaxation_side == RelaxationSide.UNDER and not self.is_rhs_convex():
            res.append(self._x)
        elif self.relaxation_side == RelaxationSide.OVER and not self.is_rhs_concave():
            res.append(self._x)
        return res

    def set_input(self, x, aux_var, shape, f_x_expr, pw_repn='INC', relaxation_side=RelaxationSide.BOTH,
                  use_linear_relaxation=True, large_coef=1e5, small_coef=1e-10, safety_tol=1e-10):
        """
        Parameters
        ----------
        x: pyomo.core.base.var._GeneralVarData
            The "x" variable in aux_var = f(x).
        aux_var: pyomo.core.base.var._GeneralVarData
            The auxillary variable replacing f(x)
        shape: FunctionShape
            Options are FunctionShape.CONVEX and FunctionShape.CONCAVE
        f_x_expr: pyomo expression
            The pyomo expression representing f(x)
        pw_repn: str
            This must be one of the valid strings for the piecewise representation to use (directly from the Piecewise
            component). Use help(Piecewise) to learn more.
        relaxation_side: RelaxationSide
            Provide the desired side for the relaxation (OVER, UNDER, or BOTH)
        use_linear_relaxation: bool
            Specifies whether a linear or nonlinear relaxation should be used
        """
        super().set_input(relaxation_side=relaxation_side,
                          use_linear_relaxation=use_linear_relaxation,
                          large_coef=large_coef, small_coef=small_coef,
                          safety_tol=safety_tol)
        self._pw_repn = pw_repn
        self._function_shape = shape
        self._f_x_expr = f_x_expr

        self._xref.set_component(x)
        self._aux_var_ref.set_component(aux_var)
        bnds_list = _get_bnds_list(self._x)
        self._partitions[self._x] = bnds_list

    def build(self, x, aux_var, shape, f_x_expr, pw_repn='INC', relaxation_side=RelaxationSide.BOTH,
              use_linear_relaxation=True, large_coef=1e5, small_coef=1e-10, safety_tol=1e-10):
        """
        Parameters
        ----------
        x: pyomo.core.base.var._GeneralVarData
            The "x" variable in aux_var = f(x).
        aux_var: pyomo.core.base.var._GeneralVarData
            The auxillary variable replacing f(x)
        shape: FunctionShape
            Options are FunctionShape.CONVEX and FunctionShape.CONCAVE
        f_x_expr: pyomo expression
            The pyomo expression representing f(x)
        pw_repn: str
            This must be one of the valid strings for the piecewise representation to use (directly from the Piecewise
            component). Use help(Piecewise) to learn more.
        relaxation_side: RelaxationSide
            Provide the desired side for the relaxation (OVER, UNDER, or BOTH)
        use_linear_relaxation: bool
            Specifies whether a linear or nonlinear relaxation should be used
        """
        self.set_input(x=x, aux_var=aux_var, shape=shape, f_x_expr=f_x_expr, pw_repn=pw_repn,
                       relaxation_side=relaxation_side, use_linear_relaxation=use_linear_relaxation,
                       large_coef=large_coef, small_coef=small_coef, safety_tol=safety_tol)
        self.rebuild()

    def _remove_relaxation(self):
        del self._secant, self._secant_slope, self._secant_intercept, self._pw_secant
        self._secant = None
        self._secant_expr = None
        self._secant_slope = None
        self._secant_intercept = None
        self._pw_secant = None

    def remove_relaxation(self):
        super().remove_relaxation()
        self._remove_relaxation()

    def _needs_secant(self):
        if self.relaxation_side == RelaxationSide.BOTH and (self.is_rhs_convex() or self.is_rhs_concave()):
            return True
        elif self.relaxation_side == RelaxationSide.UNDER and self.is_rhs_concave():
            return True
        elif self.relaxation_side == RelaxationSide.OVER and self.is_rhs_convex():
            return True
        else:
            return False

    def rebuild(self, build_nonlinear_constraint=False, ensure_oa_at_vertices=True):
        super().rebuild(build_nonlinear_constraint=build_nonlinear_constraint,
                        ensure_oa_at_vertices=ensure_oa_at_vertices)
        if not build_nonlinear_constraint:
            if self._check_valid_domain_for_relaxation():
                if self._needs_secant():
                    if len(self._partitions[self._x]) == 2:
                        if self._secant is None:
                            self._remove_relaxation()
                            self._build_secant()
                        self._update_secant()
                    else:
                        self._remove_relaxation()
                        self._build_pw_secant()
                else:
                    self._remove_relaxation()
            else:
                self._remove_relaxation()

    def _build_secant(self):
        del self._secant_slope
        del self._secant_intercept
        del self._secant
        del self._secant_expr
        self._secant_slope = ScalarParam(mutable=True)
        self._secant_intercept = ScalarParam(mutable=True)
        e = LinearExpression(constant=self._secant_intercept, linear_coefs=[self._secant_slope], linear_vars=[self._x])
        self._secant_expr = e
        if self.is_rhs_concave():
            self._secant = ScalarConstraint(expr=self._aux_var >= e)
        elif self.is_rhs_convex():
            self._secant = ScalarConstraint(expr=self._aux_var <= e)
        else:
            raise RuntimeError('Function should be either convex or concave in order to build the secant')

    def _update_secant(self):
        _eval = _FxExpr(self._f_x_expr, self._x)
        assert len(self._partitions[self._x]) == 2

        try:
            x1 = self._partitions[self._x][0]
            x2 = self._partitions[self._x][1]
            if x1 == x2:
                slope = 0
                intercept = _eval(x1)
            else:
                y1 = _eval(x1)
                y2 = _eval(x2)
                slope = (y2 - y1) / (x2 - x1)
                intercept = y2 - slope*x2
            err_message = None
        except (ZeroDivisionError, OverflowError, ValueError) as e:
            slope = None
            intercept = None
            err_message = str(e)
        if err_message is not None:
            logger.debug(f'Encountered exception when adding secant for "{self._get_pprint_string()}"; Error message: {err_message}')
            self._remove_relaxation()
        else:
            self._secant_slope._value = slope
            self._secant_intercept._value = intercept
            if self.is_rhs_concave():
                rel_side = RelaxationSide.UNDER
            else:
                rel_side = RelaxationSide.OVER
            success, bad_var, bad_coef, err_msg = _check_cut(self._secant_expr, too_small=self.small_coef,
                                                             too_large=self.large_coef, relaxation_side=rel_side,
                                                             safety_tol=self.safety_tol)
            if not success:
                self._log_bad_cut(bad_var, bad_coef, err_msg)
                self._secant.deactivate()
            else:
                self._secant.activate()

    def _build_pw_secant(self):
        del self._pw_secant
        self._pw_secant = pe.Block(concrete=True)
        if self.is_rhs_convex():
            _pw_univariate_relaxation(b=self._pw_secant, x=self._x, w=self._aux_var, x_pts=self._partitions[self._x],
                                      f_x_expr=self._f_x_expr, pw_repn=self._pw_repn, shape=FunctionShape.CONVEX,
                                      relaxation_side=RelaxationSide.OVER, large_eval_tol=self.large_coef,
                                      safety_tol=self.safety_tol)
        else:
            _pw_univariate_relaxation(b=self._pw_secant, x=self._x, w=self._aux_var, x_pts=self._partitions[self._x],
                                      f_x_expr=self._f_x_expr, pw_repn=self._pw_repn, shape=FunctionShape.CONCAVE,
                                      relaxation_side=RelaxationSide.UNDER, large_eval_tol=self.large_coef,
                                      safety_tol=self.safety_tol)

    def add_partition_point(self, value=None):
        """
        This method adds one point to the partitioning of x. If value is not
        specified, a single point will be added to the partitioning of x at the current value of x. If value is
        specified, then value is added to the partitioning of x.

        Parameters
        ----------
        value: float
            The point to be added to the partitioning of x.
        """
        self._add_partition_point(self._x, value)

    def is_rhs_convex(self):
        """
        Returns True if linear underestimators do not need binaries. Otherwise, returns False.

        Returns
        -------
        bool
        """
        return self._function_shape == FunctionShape.CONVEX

    def is_rhs_concave(self):
        """
        Returns True if linear overestimators do not need binaries. Otherwise, returns False.

        Returns
        -------
        bool
        """
        return self._function_shape == FunctionShape.CONCAVE

    @property
    def use_linear_relaxation(self):
        return self._use_linear_relaxation

    @use_linear_relaxation.setter
    def use_linear_relaxation(self, val):
        self._use_linear_relaxation = val


@declare_custom_block(name='CustomUnivariateBaseRelaxation')
class CustomUnivariateBaseRelaxationData(PWUnivariateRelaxationData):
    """
    A helper class for building and modifying piecewise relaxations of aux_var = x**2.
    """

    def _rhs_func(self, x):
        raise NotImplementedError('This should be implemented by a derived class')

    def set_input(self, x, aux_var, pw_repn='INC', relaxation_side=RelaxationSide.BOTH,
                  use_linear_relaxation=True, large_coef=1e5, small_coef=1e-10, safety_tol=1e-10):
        """
        Parameters
        ----------
        x: pyomo.core.base.var._GeneralVarData
            The "x" variable in aux_var = f(x).
        aux_var: pyomo.core.base.var._GeneralVarData
            The auxillary variable replacing f(x)
        pw_repn: str
            This must be one of the valid strings for the piecewise representation to use (directly from the Piecewise
            component). Use help(Piecewise) to learn more.
        relaxation_side: RelaxationSide
            Provide the desired side for the relaxation (OVER, UNDER, or BOTH)
        use_linear_relaxation: bool
            Specifies whether a linear or nonlinear relaxation should be used
        """
        super().set_input(x=x, aux_var=aux_var, shape=FunctionShape.UNKNOWN,
                          f_x_expr=self._rhs_func(x), pw_repn=pw_repn,
                          relaxation_side=relaxation_side,
                          use_linear_relaxation=use_linear_relaxation,
                          large_coef=large_coef, small_coef=small_coef,
                          safety_tol=safety_tol)

    def build(self, x, aux_var, pw_repn='INC', relaxation_side=RelaxationSide.BOTH,
              use_linear_relaxation=True, large_coef=1e5, small_coef=1e-10, safety_tol=1e-10):
        """
        Parameters
        ----------
        x: pyomo.core.base.var._GeneralVarData
            The "x" variable in aux_var = f(x).
        aux_var: pyomo.core.base.var._GeneralVarData
            The auxillary variable replacing f(x)
        pw_repn: str
            This must be one of the valid strings for the piecewise representation to use (directly from the Piecewise
            component). Use help(Piecewise) to learn more.
        relaxation_side: RelaxationSide
            Provide the desired side for the relaxation (OVER, UNDER, or BOTH)
        use_linear_relaxation: bool
            Specifies whether a linear or nonlinear relaxation should be used
        """
        self.set_input(x=x, aux_var=aux_var, pw_repn=pw_repn, relaxation_side=relaxation_side,
                       use_linear_relaxation=use_linear_relaxation, large_coef=large_coef, small_coef=small_coef,
                       safety_tol=safety_tol)
        self.rebuild()


@declare_custom_block(name='PWXSquaredRelaxation')
class PWXSquaredRelaxationData(CustomUnivariateBaseRelaxationData):
    """
    A helper class for building and modifying piecewise relaxations of aux_var = x**2.
    """
    def _rhs_func(self, x):
        return x**2

    def is_rhs_convex(self):
        return True


@declare_custom_block(name='PWCosRelaxation')
class PWCosRelaxationData(CustomUnivariateBaseRelaxationData):
    """
    A helper class for building and modifying piecewise relaxations of w = cos(x) for -pi/2 <= x <= pi/2.
    """

    def __init__(self, component):
        super().__init__(component)
        self._last_concave = None

    def rebuild(self, build_nonlinear_constraint=False, ensure_oa_at_vertices=True):
        current_concave = self.is_rhs_concave()
        if current_concave != self._last_concave:
            self._needs_rebuilt = True
        self._last_concave = current_concave
        super().rebuild(build_nonlinear_constraint=build_nonlinear_constraint,
                        ensure_oa_at_vertices=ensure_oa_at_vertices)

    def _rhs_func(self, x):
        return pe.cos(x)

    def is_rhs_concave(self):
        lb, ub = tuple(_get_bnds_list(self._x))
        if lb >= -math.pi/2 and ub <= math.pi/2:
            return True
        else:
            return False


@declare_custom_block(name='SinArctanBaseRelaxation')
class SinArctanBaseRelaxationData(CustomUnivariateBaseRelaxationData):
    """
    A helper class for building and modifying piecewise relaxations of w = sin(x) for -pi/2 <= x <= pi/2.
    """

    def _rhs_func(self, x):
        raise NotImplementedError('This should be implemented by a derived class')

    def __init__(self, component):
        super().__init__(component)
        self._secant_index = None
        self._secant_exprs = None
        self._last_convex = None
        self._last_concave = None

    def _remove_relaxation(self):
        super()._remove_relaxation()
        del self._secant_index
        del self._secant_exprs
        self._secant_index = None
        self._secant_exprs = None

    def _pw_func(self):
        raise NotImplementedError('This should be implemented by a derived class')

    def _underestimator_func(self):
        raise NotImplementedError('This should be implemented by a derived class')

    def _overestimator_func(self):
        raise NotImplementedError('This should be implemented by a derived class')

    def rebuild(self, build_nonlinear_constraint=False, ensure_oa_at_vertices=True):
        current_convex = self.is_rhs_convex()
        current_concave = self.is_rhs_concave()
        if current_convex != self._last_convex or current_concave != self._last_concave:
            self._needs_rebuilt = True
        self._last_convex = current_convex
        self._last_concave = current_concave
        super().rebuild(build_nonlinear_constraint=build_nonlinear_constraint,
                        ensure_oa_at_vertices=ensure_oa_at_vertices)
        if not build_nonlinear_constraint:
            if self._check_valid_domain_for_relaxation():
                if (not self.is_rhs_convex()) and (not self.is_rhs_concave()):
                    if len(self._partitions[self._x]) == 2:
                        if self._secant is None:
                            self._remove_relaxation()
                            self._build_relaxation()
                        self._update_relaxation()
                    else:
                        self._remove_relaxation()
                        del self._pw_secant
                        self._pw_secant = pe.Block(concrete=True)
                        self._pw_func()(b=self._pw_secant, x=self._x, w=self._aux_var, x_pts=self._partitions[self._x],
                                        relaxation_side=self.relaxation_side,
                                        safety_tol=self.safety_tol)
            else:
                self._remove_relaxation()

    def _build_relaxation(self):
        del self._secant_index, self._secant_slope, self._secant_intercept, self._secant
        self._secant_index = pe.Set(initialize=[0, 1, 2, 3])
        self._secant_exprs = dict()
        self._secant_slope = IndexedParam(self._secant_index, mutable=True)
        self._secant_intercept = IndexedParam(self._secant_index, mutable=True)
        self._secant = IndexedConstraint(self._secant_index)
        if self.relaxation_side in {RelaxationSide.BOTH, RelaxationSide.UNDER}:
            for ndx in [0, 1]:
                e = LinearExpression(constant=self._secant_intercept[ndx],
                                     linear_coefs=[self._secant_slope[ndx]], linear_vars=[self._x])
                self._secant_exprs[ndx] = e
                self._secant[ndx] = self._aux_var >= e
        if self.relaxation_side in {RelaxationSide.BOTH, RelaxationSide.OVER}:
            for ndx in [2, 3]:
                e = LinearExpression(constant=self._secant_intercept[ndx],
                                     linear_coefs=[self._secant_slope[ndx]], linear_vars=[self._x])
                self._secant_exprs[ndx] = e
                self._secant[ndx] = self._aux_var <= e

    def _check_expr(self, ndx):
        if ndx in {0, 1}:
            rel_side = RelaxationSide.UNDER
        else:
            rel_side = RelaxationSide.OVER
        success, bad_var, bad_coef, err_msg = _check_cut(self._secant_exprs[ndx], too_small=self.small_coef,
                                                         too_large=self.large_coef, relaxation_side=rel_side,
                                                         safety_tol=self.safety_tol)
        if not success:
            self._log_bad_cut(bad_var, bad_coef, err_msg)
            self._secant[ndx].deactivate()
        else:
            self._secant[ndx].activate()

    def _update_relaxation(self):
        xlb, xub = _get_bnds_tuple(self._x)
        _eval = _FxExpr(self.get_rhs_expr(), self._x)
        if self.relaxation_side in {RelaxationSide.BOTH, RelaxationSide.UNDER}:
            tangent_x, tangent_slope, tangent_int = self._underestimator_func()(xub)
            if tangent_x >= xlb:
                self._secant_slope[0]._value = tangent_slope
                self._secant_intercept[0]._value = tangent_int
                self._secant_slope[1]._value = _eval.deriv(xlb)
                self._secant_intercept[1]._value = _eval(xlb) - xlb * _eval.deriv(xlb)
                self._check_expr(0)
                self._check_expr(1)
            else:
                y1 = _eval(xlb)
                y2 = _eval(xub)
                slope = (y2 - y1) / (xub - xlb)
                intercept = y2 - slope * xub
                self._secant_slope[0]._value = slope
                self._secant_intercept[0]._value = intercept
                self._check_expr(0)
                self._secant[1].deactivate()

        if self.relaxation_side in {RelaxationSide.BOTH, RelaxationSide.OVER}:
            tangent_x, tangent_slope, tangent_int = self._overestimator_func()(xlb)
            if tangent_x <= xub:
                self._secant_slope[2]._value = tangent_slope
                self._secant_intercept[2]._value = tangent_int
                self._secant_slope[3]._value = _eval.deriv(xub)
                self._secant_intercept[3]._value = _eval(xub) - xub * _eval.deriv(xub)
                self._check_expr(2)
                self._check_expr(3)
            else:
                y1 = _eval(xlb)
                y2 = _eval(xub)
                slope = (y2 - y1) / (xub - xlb)
                intercept = y2 - slope * xub
                self._secant_slope[2]._value = slope
                self._secant_intercept[2]._value = intercept
                self._check_expr(2)
                self._secant[3].deactivate()


@declare_custom_block(name='PWSinRelaxation')
class PWSinRelaxationData(SinArctanBaseRelaxationData):
    """
    A helper class for building and modifying piecewise relaxations of w = sin(x) for -pi/2 <= x <= pi/2.
    """

    def _rhs_func(self, x):
        return pe.sin(x)

    def _check_valid_domain_for_relaxation(self) -> bool:
        lb, ub = _get_bnds_tuple(self._x)
        if lb >= -math.pi / 2 and ub <= math.pi / 2:
            return True
        return False

    def _pw_func(self):
        return pw_sin_relaxation

    def _underestimator_func(self):
        return _compute_sine_underestimator_tangent_point

    def _overestimator_func(self):
        return _compute_sine_overestimator_tangent_point

    def is_rhs_convex(self):
        """
        Returns True if linear underestimators do not need binaries. Otherwise, returns False.

        Returns
        -------
        bool
        """
        lb, ub = tuple(_get_bnds_list(self._x))
        if lb >= -math.pi / 2 and ub <= 0:
            return True
        return False

    def is_rhs_concave(self):
        """
        Returns True if linear overestimators do not need binaries. Otherwise, returns False.

        Returns
        -------
        bool
        """
        lb, ub = tuple(_get_bnds_list(self._x))
        if lb >= 0 and ub <= math.pi / 2:
            return True
        return False


@declare_custom_block(name='PWArctanRelaxation')
class PWArctanRelaxationData(SinArctanBaseRelaxationData):
    """
    A helper class for building and modifying piecewise relaxations of w = arctan(x).
    """

    def _rhs_func(self, x):
        return pe.atan(x)

    def _pw_func(self):
        return pw_arctan_relaxation

    def _underestimator_func(self):
        return _compute_arctan_underestimator_tangent_point

    def _overestimator_func(self):
        return _compute_arctan_overestimator_tangent_point

    def is_rhs_convex(self):
        """
        Returns True if linear underestimators do not need binaries. Otherwise, returns False.

        Returns
        -------
        bool
        """
        lb, ub = tuple(_get_bnds_list(self._x))
        if ub <= 0:
            return True
        return False

    def is_rhs_concave(self):
        """
        Returns True if linear overestimators do not need binaries. Otherwise, returns False.

        Returns
        -------
        bool
        """
        lb, ub = tuple(_get_bnds_list(self._x))
        if lb >= 0:
            return True
        return False



import pyomo.environ as pe
from pyomo.core.expr.sympy_tools import sympyify_expression, sympy2pyomo_expression
from pyomo.core.expr.numvalue import is_fixed


def get_objective(m):
    """
    Assert that there is only one active objective in m and return it.

    Parameters
    ----------
    m: pyomo.core.base.block._BlockData

    Returns
    -------
    obj: pyomo.core.base.objective._ObjectiveData
    """
    obj = None
    for o in m.component_data_objects(pe.Objective, descend_into=True, active=True, sort=True):
        if obj is not None:
            raise ValueError('Found multiple active objectives')
        obj = o
    return obj


def simplify_expr(expr):
    if is_fixed(expr):
        return pe.value(expr)
    else:
        om, se = sympyify_expression(expr)
        se = se.simplify()
        return sympy2pyomo_expression(se, om)



from .mccormick import PWMcCormickRelaxationData, PWMcCormickRelaxation
from .univariate import PWXSquaredRelaxationData, PWUnivariateRelaxationData, PWArctanRelaxationData, \
    PWCosRelaxationData, PWSinRelaxationData
from .univariate import PWXSquaredRelaxation, PWUnivariateRelaxation, PWArctanRelaxation, \
    PWCosRelaxation, PWSinRelaxation
from .alphabb import AlphaBBRelaxation, AlphaBBRelaxationData
from .multivariate import MultivariateRelaxationData, MultivariateRelaxation
from pyomo.core.expr.visitor import replace_expressions
from coramin.utils.coramin_enums import FunctionShape


def copy_relaxation_with_local_data(rel, old_var_to_new_var_map):
    """
    This function copies a relaxation object with new variables.
    Note that only what can be set through the set_input and build
    methods are copied. For example, piecewise partitioning points
    are not copied.

    Parameters
    ----------
    rel: coramin.relaxations.relaxations_base.BaseRelaxationData
        The relaxation to be copied
    old_var_to_new_var_map: dict
        Map from the original variable id to the new variable

    Returns
    -------
    rel: coramin.relaxations.relaxations_base.BaseRelaxationData
        The copy of rel with new variables
    """
    if isinstance(rel, PWXSquaredRelaxationData):
        new_x = old_var_to_new_var_map[id(rel.get_rhs_vars()[0])]
        new_aux_var = old_var_to_new_var_map[id(rel.get_aux_var())]
        new_rel = PWXSquaredRelaxation(concrete=True)
        new_rel.set_input(x=new_x, aux_var=new_aux_var, pw_repn=rel._pw_repn,
                          use_linear_relaxation=rel.use_linear_relaxation,
                          relaxation_side=rel.relaxation_side)
    elif isinstance(rel, PWArctanRelaxationData):
        new_x = old_var_to_new_var_map[id(rel.get_rhs_vars()[0])]
        new_aux_var = old_var_to_new_var_map[id(rel.get_aux_var())]
        new_rel = PWArctanRelaxation(concrete=True)
        new_rel.set_input(x=new_x, aux_var=new_aux_var, pw_repn=rel._pw_repn,
                          relaxation_side=rel.relaxation_side,
                          use_linear_relaxation=rel.use_linear_relaxation)
    elif isinstance(rel, PWSinRelaxationData):
        new_x = old_var_to_new_var_map[id(rel.get_rhs_vars()[0])]
        new_aux_var = old_var_to_new_var_map[id(rel.get_aux_var())]
        new_rel = PWSinRelaxation(concrete=True)
        new_rel.set_input(x=new_x, aux_var=new_aux_var, pw_repn=rel._pw_repn,
                          relaxation_side=rel.relaxation_side,
                          use_linear_relaxation=rel.use_linear_relaxation)
    elif isinstance(rel, PWCosRelaxationData):
        new_x = old_var_to_new_var_map[id(rel.get_rhs_vars()[0])]
        new_aux_var = old_var_to_new_var_map[id(rel.get_aux_var())]
        new_rel = PWCosRelaxation(concrete=True)
        new_rel.set_input(x=new_x, aux_var=new_aux_var, pw_repn=rel._pw_repn,
                          relaxation_side=rel.relaxation_side,
                          use_linear_relaxation=rel.use_linear_relaxation)
    elif isinstance(rel, PWUnivariateRelaxationData):
        new_x = old_var_to_new_var_map[id(rel.get_rhs_vars()[0])]
        new_aux_var = old_var_to_new_var_map[id(rel.get_aux_var())]
        new_f_x_expr = replace_expressions(rel.get_rhs_expr(),
                                           substitution_map=old_var_to_new_var_map,
                                           remove_named_expressions=True)
        new_rel = PWUnivariateRelaxation(concrete=True)
        if rel.is_rhs_convex():
            shape = FunctionShape.CONVEX
        elif rel.is_rhs_concave():
            shape = FunctionShape.CONCAVE
        else:
            shape = FunctionShape.UNKNOWN
        new_rel.set_input(x=new_x, aux_var=new_aux_var, shape=shape,
                          f_x_expr=new_f_x_expr, pw_repn=rel._pw_repn,
                          relaxation_side=rel.relaxation_side,
                          use_linear_relaxation=rel.use_linear_relaxation)
    elif isinstance(rel, PWMcCormickRelaxationData):
        rhs_vars = rel.get_rhs_vars()
        old_x1 = rhs_vars[0]
        old_x2 = rhs_vars[1]
        new_x1 = old_var_to_new_var_map[id(old_x1)]
        new_x2 = old_var_to_new_var_map[id(old_x2)]
        new_aux_var = old_var_to_new_var_map[id(rel.get_aux_var())]
        new_rel = PWMcCormickRelaxation(concrete=True)
        new_rel.set_input(x1=new_x1, x2=new_x2, aux_var=new_aux_var,
                          relaxation_side=rel.relaxation_side)
    elif isinstance(rel, AlphaBBRelaxationData):
        new_aux_var = old_var_to_new_var_map[id(rel.get_aux_var())]
        new_f_x_expr = replace_expressions(
            rel.get_rhs_expr(),
            substitution_map=old_var_to_new_var_map,
            remove_named_expressions=True
        )
        new_rel = AlphaBBRelaxation(concrete=True)
        new_rel.set_input(
            aux_var=new_aux_var,
            f_x_expr=new_f_x_expr,
            relaxation_side=rel.relaxation_side,
            use_linear_relaxation=rel.use_linear_relaxation,
            eigenvalue_bounder=rel.hessian.method,
            eigenvalue_opt=rel.hessian.opt,
        )
    elif isinstance(rel, MultivariateRelaxationData):
        new_aux_var = old_var_to_new_var_map[id(rel.get_aux_var())]
        if rel.is_rhs_convex():
            shape = FunctionShape.CONVEX
        elif rel.is_rhs_concave():
            shape = FunctionShape.CONCAVE
        else:
            shape = FunctionShape.UNKNOWN
        new_f_x_expr = replace_expressions(rel.get_rhs_expr(),
                                           substitution_map=old_var_to_new_var_map,
                                           remove_named_expressions=True)
        new_rel = MultivariateRelaxation(concrete=True)
        new_rel.set_input(aux_var=new_aux_var, shape=shape, f_x_expr=new_f_x_expr,
                          use_linear_relaxation=rel.use_linear_relaxation)
    else:
        raise ValueError('Unrecognized relaxation: {0}'.format(str(type(rel))))

    new_rel.small_coef = rel.small_coef
    new_rel.large_coef = rel.large_coef
    new_rel.safety_tol = rel.safety_tol

    return new_rel



from pyomo.contrib import appsi
from pyomo.common.collections import ComponentSet
from coramin.utils.coramin_enums import RelaxationSide
import pyomo.environ as pe
import time
from pyomo.common.config import ConfigValue, NonNegativeInt, NonNegativeFloat, In
from pyomo.common.timing import HierarchicalTimer
from pyomo.core.staleflag import StaleFlagManager
from coramin.utils import get_objective
from coramin.relaxations import relaxation_data_objects
from typing import Optional
from typing import Sequence, Mapping, MutableMapping, Tuple, List
from pyomo.core.base.var import _GeneralVarData
from pyomo.core.base.constraint import _GeneralConstraintData
from pyomo.core.base.param import _ParamData


import logging

logger = logging.getLogger(__name__)


class ECPConfig(appsi.base.SolverConfig):
    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super(ECPConfig, self).__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        self.feasibility_tol = self.declare(
            "feasibility_tol",
            ConfigValue(
                default=1e-6,
                domain=NonNegativeFloat,
                doc="Tolerance below which cuts will not be added",
            ),
        )
        self.max_iter = self.declare(
            "max_iter",
            ConfigValue(
                default=30, domain=NonNegativeInt, doc="Maximum number of iterations"
            ),
        )
        self.keep_cuts = self.declare(
            "keep_cuts",
            ConfigValue(
                default=False,
                domain=In([True, False]),
                doc="Whether or not to keep the cuts generated after the solve",
            ),
        )

        self.time_limit = 600


class ECPSolutionLoader(appsi.base.SolutionLoaderBase):
    def __init__(self, primals: MutableMapping[_GeneralVarData, float]):
        self._primals = primals

    def get_primals(
        self, vars_to_load: Optional[Sequence[_GeneralVarData]] = None
    ) -> Mapping[_GeneralVarData, float]:
        if vars_to_load is None:
            primals = pe.ComponentMap(self._primals)
        else:
            primals = pe.ComponentMap()
            for v in vars_to_load:
                primals[v] = self._primals[v]
        return primals


class ECPResults(appsi.base.Results):
    def __init__(self):
        super(ECPResults, self).__init__()
        self.wallclock_time = None


class ECPBounder(appsi.base.PersistentSolver):
    """
    A solver designed for use inside of OBBT. This solver is a persistent solver for
    efficient changes to the objective. Additionally, it provides a mechanism for
    refining convex nonlinear constraints during OBBT.
    """

    def __init__(self, subproblem_solver: appsi.base.PersistentSolver):
        super(ECPBounder, self).__init__()
        self._subproblem_solver = subproblem_solver
        self._relaxations = ComponentSet()
        self._relaxations_with_added_cuts = ComponentSet()
        self._pyomo_model = None
        self._config = ECPConfig()
        self._start_time: Optional[float] = None
        self._update_config = appsi.base.UpdateConfig()

    def available(self):
        return self._subproblem_solver.available()

    def version(self) -> Tuple:
        return 0, 1, 0

    @property
    def symbol_map(self):
        raise NotImplementedError("ECPBounder does not use a symbol map")

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, val: ECPConfig):
        self._config = val

    @property
    def update_config(self) -> appsi.base.UpdateConfig:
        return self._update_config

    @update_config.setter
    def update_config(self, val: appsi.base.UpdateConfig):
        self._update_config = val

    @property
    def _elapsed_time(self):
        return time.time() - self._start_time

    @property
    def _remaining_time(self):
        return max(0.0, self.config.time_limit - self._elapsed_time)

    def solve(self, model, timer: HierarchicalTimer = None) -> ECPResults:
        self._start_time = time.time()
        if timer is None:
            timer = HierarchicalTimer()
        timer.start("ECP Solve")
        StaleFlagManager.mark_all_as_stale()
        logger.info(
            "{0:<10}{1:<12}{2:<12}{3:<12}{4:<12}".format(
                "Iter", "objective", "max_viol", "time", "# cuts"
            )
        )
        self._pyomo_model = model

        obj = get_objective(model)
        if obj is None:
            raise ValueError("Could not find any active objectives")

        final_res = ECPResults()
        self._relaxations = ComponentSet()
        self._relaxations_with_added_cuts = ComponentSet()
        for b in relaxation_data_objects(
            self._pyomo_model, descend_into=True, active=True
        ):
            self._relaxations.add(b)

        self._subproblem_solver.config.load_solution = False
        orig_var_vals = pe.ComponentMap()
        for v in self._pyomo_model.component_data_objects(pe.Var, descend_into=True):
            orig_var_vals[v] = v.value

        all_added_cons = list()
        for _iter in range(self.config.max_iter):
            if self._elapsed_time >= self.config.time_limit:
                final_res.termination_condition = (
                    appsi.base.TerminationCondition.maxTimeLimit
                )
                logger.warning("ECPBounder: time limit reached.")
                break
            self._subproblem_solver.config.time_limit = self._remaining_time
            res = self._subproblem_solver.solve(self._pyomo_model, timer=timer)
            if res.termination_condition == appsi.base.TerminationCondition.optimal:
                res.solution_loader.load_vars()
            else:
                final_res.termination_condition = res.termination_condition
                logger.warning("ECPBounder: subproblem did not terminate optimally")
                break

            new_con_list = list()
            max_viol = 0
            for b in self._relaxations:
                viol = None
                try:
                    if b.is_rhs_convex() and b.relaxation_side in {
                        RelaxationSide.BOTH,
                        RelaxationSide.UNDER,
                    }:
                        viol = pe.value(b.get_rhs_expr()) - b.get_aux_var().value
                    elif b.is_rhs_concave() and b.relaxation_side in {
                        RelaxationSide.BOTH,
                        RelaxationSide.OVER,
                    }:
                        viol = b.get_aux_var().value - pe.value(b.get_rhs_expr())
                except (OverflowError, ZeroDivisionError, ValueError) as err:
                    logger.warning("could not generate ECP cut due to " + str(err))
                if viol is not None:
                    if viol > max_viol:
                        max_viol = viol
                    if viol > self.config.feasibility_tol:
                        new_con = b.add_cut(
                            keep_cut=self.config.keep_cuts,
                            check_violation=True,
                            feasibility_tol=self.config.feasibility_tol,
                        )
                        if new_con is not None:
                            self._relaxations_with_added_cuts.add(b)
                            new_con_list.append(new_con)
            self._subproblem_solver.add_constraints(new_con_list)

            final_res.best_objective_bound = res.best_objective_bound
            logger.info(
                "{0:<10d}{1:<12.3e}{2:<12.3e}{3:<12.3e}{4:<12d}".format(
                    _iter,
                    final_res.best_objective_bound,
                    max_viol,
                    self._elapsed_time,
                    len(new_con_list),
                )
            )

            all_added_cons.extend(new_con_list)

            if len(new_con_list) == 0:
                # The goal of the ECPBounder is not to find the optimal solution.
                # Rather, the goal is just to get a decent bound quickly.
                # However, if the problem is convex, we may still be able to declare
                # optimality
                final_res.termination_condition = (
                    appsi.base.TerminationCondition.unknown
                )
                logger.info("ECPBounder: converged!")

                found_feasible_solution = True
                for b in self._relaxations:
                    deviation = b.get_deviation()
                    if deviation > self.config.feasibility_tol:
                        found_feasible_solution = False

                if found_feasible_solution:
                    final_res.termination_condition = (
                        appsi.base.TerminationCondition.optimal
                    )
                    final_res.best_feasible_objective = final_res.best_objective_bound
                    primal_sol = res.solution_loader.get_primals()
                    final_res.solution_loader = ECPSolutionLoader(primal_sol)

                break

            if _iter == self.config.max_iter - 1:
                final_res.termination_condition = (
                    appsi.base.TerminationCondition.maxIterations
                )
                logger.warning("ECPBounder: reached maximum number of iterations")

        if not self.config.keep_cuts:
            self._subproblem_solver.remove_constraints(all_added_cons)
            for b in self._relaxations_with_added_cuts:
                b.rebuild()

        if final_res.termination_condition == appsi.base.TerminationCondition.optimal:
            if not self.config.load_solution:
                for v, val in orig_var_vals.items():
                    v.value = val
        else:
            if self.config.load_solution:
                raise RuntimeError(
                    "A feasible solution was not found, so no solution can be loaded. "
                    "Please set opt.config.load_solution=False and check "
                    "results.termination_condition and results.best_feasible_objective "
                    "before loading a solution."
                )
            for v, val in orig_var_vals.items():
                v.value = val

        final_res.wallclock_time = self._elapsed_time
        timer.stop("ECP Solve")
        return final_res

    def set_instance(self, model):
        saved_update_config = self.update_config
        saved_config = self.config
        self.__init__(self._subproblem_solver)
        self.config = saved_config
        self.update_config = saved_update_config
        self._pyomo_model = model
        self._subproblem_solver.set_instance(model)

    def get_primals(
        self, vars_to_load: Optional[Sequence[_GeneralVarData]] = None
    ) -> Mapping[_GeneralVarData, float]:
        return self._subproblem_solver.get_primals(vars_to_load=vars_to_load)

    def add_block(self, block):
        self._subproblem_solver.add_block(block)

    def add_constraints(self, cons: List[_GeneralConstraintData]):
        self._subproblem_solver.add_constraints(cons)

    def add_variables(self, variables: List[_GeneralVarData]):
        self._subproblem_solver.add_variables(variables=variables)

    def add_params(self, params: List[_ParamData]):
        self._subproblem_solver.add_params(params=params)

    def remove_block(self, block):
        self._subproblem_solver.remove_block(block)

    def remove_constraints(self, cons: List[_GeneralConstraintData]):
        self._subproblem_solver.remove_constraints(cons=cons)

    def remove_variables(self, variables: List[_GeneralVarData]):
        self._subproblem_solver.remove_variables(variables=variables)

    def remove_params(self, params: List[_ParamData]):
        self._subproblem_solver.remove_params(params=params)

    def set_objective(self, obj):
        self._subproblem_solver.set_objective(obj)

    def update_variables(self, variables: List[_GeneralVarData]):
        self._subproblem_solver.update_variables(variables=variables)

    def update_params(self):
        return self._subproblem_solver.update_params()



import pyomo.environ as pe
from .relaxations_base import BaseRelaxationData
from typing import Generator


def relaxation_data_objects(
    block, descend_into=True, active=None, sort=False
) -> Generator[BaseRelaxationData, None, None]:
    """
    Iterate over all instances of BaseRelaxationData in the block.

    Parameters
    ----------
    block: pyomo.core.base.block._BlockData
        The Block in which to look for relaxations
    descend_into: bool
        Whether or not to look for relaxations in sub-blocks
    active: bool
        If True, then any relaxations that have been deactivated or live on deactivated blocks will not be returned.
    sort: bool

    Returns
    -------
    relaxations: generator
        A generator yielding the relaxation objects.
    """
    for b in block.component_data_objects(
        pe.Block, descend_into=descend_into, active=active, sort=sort
    ):
        if isinstance(b, BaseRelaxationData):
            yield b


def _nonrelaxation_block_objects(block, descend_into=True, active=None, sort=False):
    for b in block.component_data_objects(
        pe.Block, descend_into=False, active=active, sort=sort
    ):
        if isinstance(b, BaseRelaxationData):
            continue
        else:
            yield b
            if descend_into:
                for _b in _nonrelaxation_block_objects(
                    b, descend_into=True, active=active, sort=sort
                ):
                    yield _b


def nonrelaxation_component_data_objects(
    block, ctype=None, active=None, sort=False, descend_into=True
):
    """
    Iterate over all components with the corresponding ctype (e.g., Constraint) in the block excluding
    those instances which are or live on relaxation objects (instances of BaseRelaxationData).

    Parameters
    ----------
    block: pyomo.core.base.block._BlockData
        The Block in which to look for components
    ctype: type
        The type of component to iterate over
    descend_into: bool
        Whether or not to look for components in sub-blocks
    active: bool
        If True, then any components that have been deactivated or live on deactivated blocks will not be returned.
    sort: bool

    Returns
    -------
    components: generator
        A generator yielding the requested components.
    """
    if not isinstance(ctype, type):
        raise ValueError(
            "nonrelaxation_component_data_objects expects ctype to be a type, not a "
            + str(type(ctype))
        )
    if ctype is pe.Block:
        for b in _nonrelaxation_block_objects(
            block, descend_into=descend_into, active=active, sort=sort
        ):
            yield b
    else:
        for comp in block.component_data_objects(
            ctype=ctype, descend_into=False, active=active, sort=sort
        ):
            yield comp
        if descend_into:
            for b in _nonrelaxation_block_objects(
                block, descend_into=True, active=active, sort=sort
            ):
                for comp in b.component_data_objects(
                    ctype=ctype, descend_into=False, active=active, sort=sort
                ):
                    yield comp
